/*
 * gui.c: implements the main window
 * Copyright (C) 2002-2004 Saulius Menkevicius
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <string.h>

#include <glib.h>
#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>

#include "main.h"
#include "prefs.h"
#include "user.h"
#include "sess.h"
#include "gui.h"
#include "gui_page.h"
#include "gui_ulist.h"
#include "gui_msg.h"
#include "gui_channel.h"
#include "gui_misc.h"
#include "gui_tray.h"
#include "gui_topic_dlg.h"
#include "gui_netselect_dlg.h"

/** GUI GEOMETRIES  &  OTHER SETTINGS
  *********************************************/
#define GEOM_NICKNAME_LEN	112

/** enums & structs
  *********************************************/

struct iface_main_wnd {
	GtkWidget
		* widget,
		* text_entry,
		* hbox1,
		* config_btn, * nickname_entry_w, * usermode_opt,
		* vsep1,
		* hpaned,
		* notebook,
		* userlist, * userlist_frame_w;

	gpointer userlist_menu_current_data;
		/* here we save user_data of activated user list item */

	gboolean topic_modified;
	gboolean has_focus;
	gboolean is_obscured;
};

/* Static variables
 ******************************/
static struct iface_main_wnd * main_wnd;
static int *tmp_p_argc;
static char *** tmp_p_argv;

/* static routines
 ******************************/

static void
usermode_selected_cb(gint new_mode, gpointer data)
{
	if(new_mode >= 0)
		prefs_set(CFG_MODE, new_mode);
}

/** update_userlist_count:
 *	updates user count on userlist frame label (including ourselves)
 */
static void
update_userlist_count()
{
	gchar * label;

	g_assert(main_wnd && main_wnd->userlist_frame_w);

	label = g_strdup_printf(_("Network users: %d"), user_count()+1);
	gtk_frame_set_label(GTK_FRAME(main_wnd->userlist_frame_w),label);
	g_free(label);
}

/**
 * main_wnd_delete_event_cb:
 *	handles window close event
 */
static gint
main_wnd_delete_event_cb(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	if(gui_tray_embedded()) {
		/* hide the window, if we have our icon on the tray */
		gui_set_visible(FALSE);
	} else {	
		raise_event(EVENT_IFACE_EXIT, NULL, 0);
	}

	return TRUE;	/* don't destroy window */
}

/**
 * main_wnd_focus_out_cb:
 *	make text entry a focused widget
 *	each time leave focus
 */
static gboolean
main_wnd_focus_out_cb(
	GtkWidget * window, GdkEventFocus * event, gpointer user_data)
{
	gtk_widget_grab_focus(main_wnd->text_entry);
	return FALSE;
}

/**
 * main_wnd_key_press_cb:
 *	handles key presses
 *	(switches pages on alt-'-', alt-'=', ctrl-pgup/down
 */
static gboolean
main_wnd_key_press_cb(GtkWidget * wnd_w, GdkEventKey * key_e, gpointer user_data)
{
	/* get keyboard modifiers */
	gpointer page;
	gboolean alt	= (key_e->state & GDK_MOD1_MASK)!=0,
		ctrl	= (key_e->state & GDK_CONTROL_MASK)!=0;

	if(key_e->keyval==GDK_Escape && gui_tray_embedded()) {
		gui_set_visible(FALSE);
		return TRUE;
	}
	else if((alt && key_e->keyval==GDK_minus) || (ctrl && key_e->keyval==GDK_Page_Up)) {
		gui_page_switch_prev();
	}
	else if((alt && key_e->keyval==GDK_equal) || (ctrl && key_e->keyval==GDK_Page_Down)) {
		gui_page_switch_next();
	}
	else if(key_e->keyval==GDK_Page_Up || key_e->keyval==GDK_Page_Down) {
		/* scrool the page up/down
		 */
		page = gui_page_current();
		if(page) {
			gui_page_scroll(
				page,
				key_e->keyval==GDK_Page_Up
					? SCROLL_PAGE_UP: SCROLL_PAGE_DOWN);
		}
	}
	else if(ctrl && (key_e->keyval==GDK_w || key_e->keyval==GDK_W)) {
		/* close active page on CTRL-W */
		raise_event(EVENT_IFACE_PAGE_CLOSE, NULL, 0);
	}
	else if(ctrl && (key_e->keyval==GDK_t || key_e->keyval==GDK_T)) {
		/* popup topic change dialog for the current session */
		if(!sess_topic_readonly(sess_current())) {
			raise_event(EVENT_IFACE_SHOW_TOPIC_DLG, sess_current(), 0);
		}
	}
	else {
		/* we're not interested: pass the event down the chain */
		return FALSE;
	}

	/* return focus to edit box */
	gtk_widget_grab_focus(main_wnd->text_entry);
	return TRUE;
}

/**
 * main_wnd_focus_change_cb:
 *	invoked when the main windows gets/looses the kbd focus
 */
static gboolean
main_wnd_focus_change_cb(
	GtkWidget * widget, GdkEventFocus * focus_event, gpointer dummy)
{
	gboolean focus_state = focus_event->in!=0;

	if(focus_state != main_wnd->has_focus) {
		main_wnd->has_focus = focus_state;

		raise_event(EVENT_IFACE_ACTIVE_CHANGE, 0, focus_state);
	}

	return FALSE;
}

static gboolean
main_wnd_visibility_notify_cb(
	GtkWidget * widget, GdkEventVisibility * event, gpointer dummy)
{
	main_wnd->is_obscured = (event->state==GDK_VISIBILITY_FULLY_OBSCURED);
	return FALSE;
}

static gboolean
entry_activate_cb(GtkWidget * widget, gpointer user)
{
	const gchar * new_nick;

	g_assert(main_wnd);
	g_assert(main_wnd->text_entry);

	if(widget==main_wnd->text_entry) {
		raise_event(
			EVENT_IFACE_TEXT_ENTER,
			(gpointer)gtk_entry_get_text(GTK_ENTRY(main_wnd->text_entry)),
			0);
		return TRUE;
	}
	else if(widget==main_wnd->nickname_entry_w) {
		new_nick = gtk_entry_get_text(GTK_ENTRY(main_wnd->nickname_entry_w));

		if(nickname_valid(new_nick)) {
			raise_event(EVENT_IFACE_NICKNAME_ENTERED, (gpointer)new_nick, 0);
		}

		gtk_widget_grab_focus(main_wnd->text_entry);
	}
	return FALSE;
}

/**
 * nickname_focus_out_cb:
 *	restores nickname entry if/when it loses focus
 */
static gboolean
nickname_focus_out_cb(GtkWidget * entry_w, GdkEventFocus * event, gpointer data)
{
	if(app_status()==APP_RUNNING) {
		gtk_entry_set_text(GTK_ENTRY(main_wnd->nickname_entry_w), my_nickname());
	}
	return FALSE;
}

static struct iface_main_wnd *
new_main_wnd()
{
	GtkWidget * w, * b;
	struct iface_main_wnd * mw;

	mw = g_malloc(sizeof(struct iface_main_wnd));

	mw->topic_modified = FALSE;
	mw->has_focus = FALSE;
	mw->is_obscured = FALSE;

	mw->widget = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gui_misc_set_icon_for(GTK_WINDOW(mw->widget));
	gtk_window_set_title(GTK_WINDOW(mw->widget), PACKAGE);
	gtk_container_set_border_width(GTK_CONTAINER(mw->widget), 2);
	g_signal_connect(G_OBJECT(mw->widget), "delete-event",
			G_CALLBACK(main_wnd_delete_event_cb), NULL);
	g_signal_connect(G_OBJECT(mw->widget), "focus-out-event",
			G_CALLBACK(main_wnd_focus_out_cb), NULL);
	g_signal_connect(G_OBJECT(mw->widget), "key-press-event",
			G_CALLBACK(main_wnd_key_press_cb), NULL);
	g_signal_connect(G_OBJECT(mw->widget), "focus-in-event",
			G_CALLBACK(main_wnd_focus_change_cb), NULL);
	g_signal_connect(G_OBJECT(mw->widget), "focus-out-event",
			G_CALLBACK(main_wnd_focus_change_cb), NULL);
	g_signal_connect(G_OBJECT(mw->widget), "visibility-notify-event",
			G_CALLBACK(main_wnd_visibility_notify_cb), NULL);

	/* create horizontal paned */
	mw->hpaned = gtk_hpaned_new();
	w = gtk_vbox_new(FALSE, 2);
	gtk_paned_pack1(GTK_PANED(mw->hpaned), w, TRUE, TRUE);
	
	/* notebook */
	mw->notebook = (GtkWidget *) gui_page_alloc_notebook();
	gtk_box_pack_start(GTK_BOX(w), mw->notebook, TRUE, TRUE, 0);

	/* text edit */
	mw->text_entry = gtk_entry_new();
	gtk_entry_set_max_length(GTK_ENTRY(mw->text_entry), MAX_TEXT_LENGTH);
	gtk_box_pack_end(GTK_BOX(w), mw->text_entry, FALSE, FALSE, 0);
	g_signal_connect(G_OBJECT(mw->text_entry), "activate",
			G_CALLBACK(entry_activate_cb), NULL);

	gtk_widget_show_all(mw->hpaned);
	gtk_container_add(GTK_CONTAINER(mw->widget), mw->hpaned);

	/* nickname btn, mode menu */
	w = gtk_vbox_new(FALSE, 2);

	/* nickname and usermode selection widgets */
	b = gtk_vbox_new(FALSE, 2);
	gtk_container_set_border_width(GTK_CONTAINER(b), 3);
	gtk_box_pack_start(GTK_BOX(w), b, FALSE, FALSE, 0);

	/* nickname entry */
	mw->nickname_entry_w = gtk_entry_new();
	gtk_entry_set_max_length(GTK_ENTRY(mw->nickname_entry_w), MAX_NICK_LENGTH);
	gtk_widget_set_usize(GTK_WIDGET(mw->nickname_entry_w), GEOM_NICKNAME_LEN, 0);
	gtk_box_pack_start(GTK_BOX(b), mw->nickname_entry_w, FALSE, FALSE, 0);
	g_signal_connect(G_OBJECT(mw->nickname_entry_w), "activate",
			G_CALLBACK(entry_activate_cb), NULL);
	g_signal_connect(G_OBJECT(mw->nickname_entry_w), "focus-out-event",
			G_CALLBACK(nickname_focus_out_cb), NULL);

	/* user mode selection menu */
	mw->usermode_opt = util_user_mode_option(usermode_selected_cb, NULL);
	gtk_box_pack_start(GTK_BOX(b), mw->usermode_opt, FALSE, FALSE, 0);

	gtk_widget_show_all(w);
	gtk_paned_pack2(GTK_PANED(mw->hpaned), w, FALSE, FALSE);

	/* user list */
	mw->userlist_frame_w = gtk_frame_new(_("Network Users"));
	b = gtk_vbox_new(FALSE, 2);
	gtk_container_set_border_width(GTK_CONTAINER(b), 3);
	gtk_container_add(GTK_CONTAINER(mw->userlist_frame_w), b);
	mw->userlist = gui_ulist_create();
	gtk_box_pack_start(GTK_BOX(b), mw->userlist, TRUE, TRUE, 0);

	gtk_widget_show_all(mw->userlist_frame_w);
	gtk_box_pack_start(GTK_BOX(w), mw->userlist_frame_w, TRUE, TRUE, 0);

	return mw;
}

static void
delete_main_wnd(struct iface_main_wnd * mw)
{
	if(!mw) return;

	gtk_widget_destroy(mw->widget);

	gui_ulist_free();

	g_free(mw);
}

/**
 * tray_toggle_visible:
 *	invoked when user clicks systray (docklet) icon
 */
static void
tray_toggle_visible()
{
	if(GTK_WIDGET_VISIBLE(main_wnd->widget)) {
		if(gdk_window_get_state(
				GTK_WIDGET(main_wnd->widget)->window) & GDK_WINDOW_STATE_ICONIFIED
			|| main_wnd->is_obscured)
		{
			gtk_window_present(gui_get_main_window());
		} else {
			gui_set_visible(FALSE);
		}
	} else {
		gui_set_visible(TRUE);
	}
}

/* save_window_size:
 *	stores window size in `CFG_WND_SIZE'
 */
static void
save_window_size(struct iface_main_wnd * mw)
{
	gint width, height;
	gchar * size;

	g_assert(mw && mw->widget);

	gtk_window_get_size(GTK_WINDOW(mw->widget), &width, &height);
	size = g_strdup_printf("%dx%d", width, height);
	prefs_set(CFG_WND_SIZE, size);
	g_free(size);
}

/* restore_window_size:
 *	restores window size from `CFG_WND_SIZE'
 */
static void
restore_window_size(struct iface_main_wnd * mw)
{
	int width, height;
	const gchar * size;

	g_assert(mw && mw->widget);

	/* try to parse the string in CFG_WND_SIZE */
	size = config_get_s(CFG_WND_SIZE);
	if(size) {
		if(sscanf(size, "%dx%d", &width, &height)==2) {
			if(width>0 && height>0 && !(width & ~0xffff) && !(height & ~0xffff)) {
				gtk_window_resize(GTK_WINDOW(mw->widget), width, height);
			}
		}
	}
}

/* config_change_handler:
 *	handles configuration changes
 */
static void
config_change_handler(enum config_enum cfg)
{
	switch(cfg) {
	case CFG_MODE:
		gtk_option_menu_set_history(
			GTK_OPTION_MENU(main_wnd->usermode_opt), prefs_int(CFG_MODE));
		break;
	case CFG_NICKNAME:
		gtk_entry_set_text(GTK_ENTRY(main_wnd->nickname_entry_w), my_nickname());
		break;
	default:
		break;
	}
}

static void
main_wnd_set_title_for_session(sess_id session)
{
	GString * title;
	const gchar * topic;
	guint len;

	/* build window title */
	title = g_string_new(PACKAGE" [");

	switch(sess_type(session)) {
	case SESSTYPE_CHANNEL:
		topic = sess_topic(session);

		g_string_append_printf(title, strlen(topic) ? "#%s: ": "#%s", sess_name(session));

		/* remove tabs,newlines from the topic
		 * and keep it's length at configured value
		 */
		len = 0;
		while(*topic && len < config_get_i(CFG_WND_TITLE_MAX_LEN)) {
			switch(*topic) {
			case '\t':/* FALLTHROUGH */
			case '\n':
				/* replace newlines with spaces */
				g_string_append_c(title, ' ');
				break;
			case '\r':
				break;
			default:
				g_string_append_unichar(title, g_utf8_get_char(topic));
				break;
			}
			/* advance to next char */
			topic = g_utf8_next_char(topic);
			len ++;
		}
		if(*topic) {
			/* string was truncated: append "..." */
			g_string_append(title, "...");
		}
		break;
	case SESSTYPE_PRIVATE:
		g_string_append_printf(title, _("Private chat with %s"), sess_name(session));
		break;
	case SESSTYPE_STATUS:
		g_string_append(title, _("Status"));
		break;
	}
	g_string_append_c(title, ']');

	gtk_window_set_title(GTK_WINDOW(main_wnd->widget), title->str);
	g_string_free(title, TRUE);
}

static void
gui_event_cb(enum app_event_enum e, void * p, int i)
{
	static gboolean already_got_embedded = FALSE;
	
	switch(e) {
	case EVENT_MAIN_INIT:
		g_assert(tmp_p_argc && tmp_p_argv);

		gtk_init(tmp_p_argc, tmp_p_argv);

		gui_misc_init();

		main_wnd = new_main_wnd();
		break;

	case EVENT_MAIN_START:
		if(config_get_i(CFG_KEEP_WND_SIZE)) {
			/* restore window size */
			restore_window_size(main_wnd);
		}

		gui_set_visible(TRUE);
		break;

	case EVENT_MAIN_PRECLOSE:
		if(config_get_i(CFG_KEEP_WND_SIZE)) {
			/* save window size on exit
			 */
			save_window_size(main_wnd);
		}
		break;

	case EVENT_MAIN_CLOSE:
		delete_main_wnd(main_wnd);
		main_wnd = NULL;

		gui_misc_destroy();
		break;

	case EVENT_CONFIG_CHANGED:
		config_change_handler((enum config_enum)i);
		break;

	case EVENT_CMDPROC_SET_TEXT:
		gtk_entry_set_text(GTK_ENTRY(main_wnd->text_entry), (const char*)p);
		break;

	case EVENT_USER_NEW:
	case EVENT_USER_REMOVED:
		update_userlist_count();
		break;

	case EVENT_SESSION_OPENED:
		if(config_get_i(CFG_PRESENT_ON_PRIVATE) && app_status()==APP_RUNNING) {
			/* popup the main window and select page
			 * of the new chat
			 */
			gui_set_visible(TRUE);
			sess_switch_to((sess_id)p);
		}
		break;

	case EVENT_SESSION_RENAMED:
	case EVENT_SESSION_TOPIC_CHANGED:
		if(EVENT_V(p, 0)==sess_current()) {
			main_wnd_set_title_for_session(EVENT_V(p, 0));
		}
		break;

	case EVENT_IFACE_PAGE_SWITCH:
		main_wnd_set_title_for_session(sess_current());
		break;

	case EVENT_IFACE_PAGE_RELEASE_FOCUS:
		gtk_widget_grab_focus(main_wnd->text_entry);
		break;

	case EVENT_IFACE_TRAY_CLICK:
		tray_toggle_visible();
		break;

	case EVENT_IFACE_TRAY_EMBEDDED:
		/* hide the main window if the tray got embedded
		 * (usualy on startup.. do this once, as the system tray area
		 * may reappear multiple times during our runtime)
		 */
		if(prefs_int(CFG_TRAY_HIDE_ON_STARTUP)
				&& prefs_int(PREF_NET_CONFIGURED)
				&& !already_got_embedded) {
			already_got_embedded = TRUE;
			gui_set_visible(FALSE);
		}
		break;

	case EVENT_IFACE_TRAY_DESTROYED:
		/* we've lost the tray: show the main window */
		gui_set_visible(TRUE);
		break;

	default:
		break;
	}
}

/* Exported routines
 ******************************/

void gui_register(int * p_argc, char *** p_argv)
{	
	tmp_p_argc = p_argc;
	tmp_p_argv = p_argv;

	register_event_cb(
		gui_event_cb,
		EVENT_MAIN|EVENT_CONFIG|EVENT_CMDPROC|EVENT_USER|EVENT_SESSION|EVENT_IFACE);

	/* register neighbour callbacks */
	gui_page_register();
	gui_msg_register();
	gui_channel_register();
	gui_tray_register();
	gui_ulist_register();
	gui_topic_dlg_register();
	gui_netselect_dlg_register();
}

void gui_run()
{
	gtk_main();
}

void gui_shutdown()
{
	gtk_main_quit();
}

gboolean gui_is_visible()
{
	return GTK_WIDGET_VISIBLE(main_wnd->widget);
}

gboolean gui_is_active()
{
	return main_wnd->has_focus;
}

/* gui_set_visible:
 *	shows/hides main window
 *
 *	presents the window (moves it to top of the desktop window stack) if
 *	the window is already shown and @visible==TRUE
 */
void gui_set_visible(gboolean visible)
{
	/* TODO: hide dialogs if visible==FALSE
	 */

	/* save window pos during hide/show */
	static gint prev_x = 0, prev_y = 0;

	if(!GTK_WIDGET_VISIBLE(main_wnd->widget) == !visible) {
		/* already shown/hidden */
		if(visible) {
			/* present */
			gtk_window_present(GTK_WINDOW(main_wnd->widget));
		}
		return;
	}

	if(visible) {
		gtk_widget_show(main_wnd->widget);
		gtk_window_move((GtkWindow*)main_wnd->widget, prev_x, prev_y);
		gtk_widget_grab_focus(main_wnd->text_entry);
		gtk_window_present(GTK_WINDOW(main_wnd->widget));

		raise_event(EVENT_IFACE_MAINWND_SHOWN, 0, 0);
	} else {
		gtk_window_get_position((GtkWindow*)main_wnd->widget, &prev_x, &prev_y);
		gtk_widget_hide(main_wnd->widget);

		raise_event(EVENT_IFACE_MAINWND_HIDDEN, 0, 0);
	}
}

gpointer
gui_get_main_window()
{
	return main_wnd->widget;
}

