/*
 * gui_config.c: implements configuration dialog
 * Copyright (C) 2002-2004 Saulius Menkevicius
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>

#include <glib.h>
#include <gtk/gtk.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gdk/gdkkeysyms.h>

#include "main.h"
#include "gui.h"
#include "prefs.h"
#include "net.h"
#include "gui_config.h"
#include "gui_misc.h"
#include "gui_tray.h"

#define WITH_ENCODING_COMBO_WIDGET \
	((GTK_MAJOR_VERSION > 2) || ((GTK_MAJOR_VERSION==2) && (GTK_MINOR_VERSION>=4)))

/** geometries
  ************************************/
#define GEOM_DLG_PAGE_BORDER	6

/** structure definitions
  ************************************/
struct config_dlg {
	gboolean is_changed, is_shown, gathering;

	GtkWidget * dlg_w,

		/* Network widgets */
		* net_type_w,
		* net_port_w,
		* net_reply_w,
		* net_log_w,
		* net_channel_greet_w,
		* net_charset_w,
#if(WITH_ENCODING_COMBO_WIDGET)
		* net_charset_combo_w,
#endif

		/* User list */
		* userlist_refresh_w,
		* userlist_keep_unreplied_users_w,

		/* net mask entries */
		* net_m1_label_w, * net_m1_w,
		* net_m2_label_w, * net_m2_w,

		/* Auto away widgets */
		* auto_away_w,
		* auto_offline_w,

		/* Pages */
		* pages_present_on_private,
		* pages_log_current_w,
		* date_channel_w,
		* date_private_w,
		* date_status_w,
		* pages_notebook_pos_w,

		/* Messages */
		* msg_max_w,
		* msg_max_label_w,
		* msg_show_window_w,
		* msg_ignore_mass_w,

		/* Tray icon */
		* tray_enable_w,
		* tray_hide_on_startup_w,
		* tray_trigger_joinleave_w,
		* tray_trigger_channel_w,
		* tray_trigger_private_w,
		* tray_trigger_status_w,
		* tray_trigger_topic_w,
		* tray_tooltip_lines_w,
		* tray_tooltip_lines_label_w,

		/* Misc */
		* save_on_exit_w,
		* keep_window_size_w,
		* motd_w,
		* motd_away_w,

		/* action_area buttons */
		* save_w,
		* apply_w;
};
#define PCONFIG_DLG(p) ((struct config_dlg*)p)

/** static variables
 */
static struct config_dlg * the_dlg; 

#if(WITH_ENCODING_COMBO_WIDGET)
const gchar *
supported_encodings[] = {
	"KOI8-R", "CP1251",	// cyrillic
	"ISO-8859-2",		// central europe
	"ISO-8859-13",		// baltic
	"ISO-8859-7",		// greek
	"ISO-8859-8",		// hebrew
	"ISO-8859-9",		// turkish
	"ISO-8859-15",		// western europe
	"ISO-2022-JP", "SJIS",	// japanese
	"CP1256",		// arabic
	"GB18030",		// chinese
	"UTF-8",
	NULL
};
#endif

/** static rountines
 */
static void set_config_dlg_modified(struct config_dlg *, gboolean);
static void show_config_dlg(struct config_dlg *);
static void hide_config_dlg(struct config_dlg *);

static void sync_dlg_from_config(struct config_dlg * dlg);
static void set_config_from_dlg(struct config_dlg * dlg);

static void
cfg_toggle_cb(GtkToggleButton * w, gpointer e)
{
	struct config_dlg * dlg = the_dlg;
	int set;

	set_config_dlg_modified(dlg, TRUE);

#define ENABLE_TOGGLE(t, e) gtk_widget_set_sensitive((t), (e))

	switch((enum config_enum)GPOINTER_TO_INT(e)) {
	case CFG_MSG_WINDOW:
		set = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(w));

		ENABLE_TOGGLE(dlg->msg_max_label_w, set);
		ENABLE_TOGGLE(dlg->msg_max_w, set);
		break;

	case CFG_TRAY_ICON:
		set = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(w));
		
		ENABLE_TOGGLE(dlg->tray_hide_on_startup_w, set);
		ENABLE_TOGGLE(dlg->tray_trigger_joinleave_w, set);
		ENABLE_TOGGLE(dlg->tray_trigger_topic_w, set);
		ENABLE_TOGGLE(dlg->tray_trigger_status_w, set);
		ENABLE_TOGGLE(dlg->tray_trigger_private_w, set);
		ENABLE_TOGGLE(dlg->tray_trigger_channel_w, set);
		ENABLE_TOGGLE(dlg->tray_tooltip_lines_w, set);
		ENABLE_TOGGLE(dlg->tray_tooltip_lines_label_w, set);
		break;
	default:
		break;
	}
#undef ENABLE_TOGGLE
}

static void
set_ipv4_entry(GtkWidget * entry_w, unsigned long addr)
{
	char str[32];
	
	if(addr) {
		sprintf(str, "%u.%u.%u.%u",
			(unsigned)(addr >> 24), (unsigned)((addr >> 16) & 0xff),
			(unsigned)((addr >> 8) & 0xff),
			(unsigned)(addr & 0xff));
	} else {
		str[0]='\0';
	}
	gtk_entry_set_text(GTK_ENTRY(entry_w), str);
}

static unsigned long
get_ipv4_entry(
	GtkWidget * entry_w,
	gboolean * p_ok)
{
	unsigned v[4];
	int num_read;
	const char * text = gtk_entry_get_text(GTK_ENTRY(entry_w));

	num_read = sscanf(text, "%d.%d.%d.%d", &v[0], &v[1], &v[2], &v[3]);
	
	if(num_read<4) {
		if(p_ok) {
			*p_ok = (num_read<=0) ? TRUE: FALSE;
		}
		return 0UL;
	}

	if((v[0]|v[1]|v[2]|v[3])&~0xff) {
		if(p_ok) {
			*p_ok = FALSE;
		}
		return 0UL;
	}

	if(p_ok) *p_ok = TRUE;
	return v[0]<<24 | v[1]<<16 | v[2]<<8 | v[3];
}

static GtkWidget *
add_check_button(
	GtkWidget * b,
	enum config_enum e,
	const char * title)
{
	GtkWidget * w;

	w = gtk_check_button_new_with_mnemonic(title);
	gtk_box_pack_start(GTK_BOX(b), w, FALSE, FALSE, 0);
	g_signal_connect(G_OBJECT(w), "toggled", G_CALLBACK(cfg_toggle_cb), GINT_TO_POINTER(e));

	return w;
}

static void
cfg_spin_cb(GtkSpinButton * w, gpointer cfg)
{
	struct config_dlg * dlg = the_dlg;
	int caway, coffline;

	set_config_dlg_modified(dlg, TRUE);

	switch((enum config_enum)GPOINTER_TO_INT(cfg))
	{
		/** ensure, that AUTO-AWAY <= AUTO-OFFLINE */
	case CFG_AUTO_AWAY:
		coffline = (int)gtk_spin_button_get_value(
				GTK_SPIN_BUTTON(dlg->auto_offline_w));
		caway = (int)gtk_spin_button_get_value(
				GTK_SPIN_BUTTON(dlg->auto_away_w));

		if(caway>coffline) {
			gtk_spin_button_set_value(
				GTK_SPIN_BUTTON(dlg->auto_offline_w),
				caway);
		}
		break;
	case CFG_AUTO_OFFLINE:
		coffline = (int)gtk_spin_button_get_value(
				GTK_SPIN_BUTTON(dlg->auto_offline_w));
		caway = (int)gtk_spin_button_get_value(
				GTK_SPIN_BUTTON(dlg->auto_away_w));
		
		if(coffline<caway) {
			gtk_spin_button_set_value(
				GTK_SPIN_BUTTON(dlg->auto_away_w),
				coffline);
		}
		break;
	default:
		break;	/* nothing */
	}
}

static GtkWidget *
add_spin_button(
	struct config_dlg * dlg,
	GtkWidget * b,
	enum config_enum e,
	const char * title,
	int min, int max,
	GtkWidget ** p_label_w)
{
	GtkWidget * w, * hb, * label_w;

	g_assert(dlg && b && title && min<max);

	hb = gtk_hbox_new(FALSE, 2);
	gtk_box_pack_start(GTK_BOX(b), hb, FALSE, FALSE, 0);

	label_w = gtk_label_new_with_mnemonic(title);
	gtk_box_pack_start(GTK_BOX(hb), label_w, FALSE, FALSE, 0);
	if(p_label_w) *p_label_w = label_w;

	w = gtk_spin_button_new_with_range(min, max, 1);
	gtk_label_set_mnemonic_widget(GTK_LABEL(label_w), w);
	gtk_box_pack_start(GTK_BOX(hb), w, FALSE, FALSE, 0);
	g_signal_connect(
		G_OBJECT(w), "value-changed",
		G_CALLBACK(cfg_spin_cb), GINT_TO_POINTER(e));

	return w;
}

static void
option_changed_cb(GtkOptionMenu * o, struct config_dlg * dlg)
{
	set_config_dlg_modified(dlg, TRUE);
}

static GtkWidget *
add_option(
	struct config_dlg * dlg,
	GtkWidget * b,
	enum config_enum cfg,
	const char * title,
	GList * entries,
	GtkWidget ** p_label_w)
{
	GtkWidget * o, * hb, * menu_w, * label_w;

	g_assert(b && title && entries);

	hb = gtk_hbox_new(FALSE, 2);
	gtk_box_pack_start(GTK_BOX(b), hb, FALSE, FALSE, 0);

	label_w = gtk_label_new_with_mnemonic(title);
	gtk_box_pack_start(GTK_BOX(hb), label_w, FALSE, FALSE, 0);
	if(p_label_w) *p_label_w = label_w;

	/* make menu */

	/* make option menu */
	o = gtk_option_menu_new();
	gtk_label_set_mnemonic_widget(GTK_LABEL(label_w), o);
	menu_w = gtk_menu_new();
	for(; entries; entries = entries->next) {
		gtk_menu_shell_append(
			GTK_MENU_SHELL(menu_w),
			gtk_menu_item_new_with_mnemonic((gchar*)entries->data));
	}
	gtk_option_menu_set_menu(GTK_OPTION_MENU(o), menu_w);
	gtk_box_pack_start(GTK_BOX(hb), o, FALSE, FALSE, 0);
	g_signal_connect(G_OBJECT(o), "changed", G_CALLBACK(option_changed_cb), dlg);

	return o;
}

static void
entry_changed_cb(GtkEntry * w, struct config_dlg * dlg)
{
	set_config_dlg_modified(dlg, TRUE);
}

static GtkWidget *
add_entry(
	struct config_dlg * dlg,
	GtkWidget * b,
	enum config_enum cfg,
	const char * title,
	GtkWidget ** p_label_w)
{
	GtkWidget * hb, * w, * label_w;

	hb = gtk_hbox_new(FALSE, 2);
	gtk_box_pack_start(GTK_BOX(b), hb, FALSE, FALSE, 0);

	/* make label */
	label_w = gtk_label_new_with_mnemonic(title);
	gtk_box_pack_start(GTK_BOX(hb), label_w, FALSE, FALSE, 0);
	if(p_label_w) *p_label_w = label_w;

	/* make edit entry */
	w = gtk_entry_new();
	gtk_label_set_mnemonic_widget(GTK_LABEL(label_w), w);
	gtk_box_pack_start(GTK_BOX(hb), w, TRUE, TRUE, 0);
	g_signal_connect(G_OBJECT(w), "changed", G_CALLBACK(entry_changed_cb), (gpointer)dlg);

	return w;
}

static GtkWidget *
add_widget(
	struct config_dlg * dlg,
	GtkWidget * frame_vbox,
	const gchar * label_text, GtkWidget * widget)
{
	GtkWidget * hbox = gtk_hbox_new(FALSE, 2);

	if(label_text) {
		gtk_box_pack_start(
			GTK_BOX(hbox), gtk_label_new_with_mnemonic(label_text),
			FALSE, FALSE, 0);
	}
	gtk_box_pack_start(GTK_BOX(hbox), widget, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(frame_vbox), hbox, FALSE, FALSE, 0);

	return widget;
}

/* gtk 2.4+ has support for encoding combo widget
 */
#if(WITH_ENCODING_COMBO_WIDGET)

static GtkWidget *
net_frame_encoding_combo(
	struct config_dlg * dlg,
	const gchar * label_text)
{
	GtkWidget * combo_w;
	const gchar ** encoding;
	gboolean have_current_encoding;

	combo_w = gtk_combo_box_entry_new_text();

	/* add encodings to the combo box list */
	have_current_encoding = FALSE;
	for(encoding = supported_encodings; *encoding; encoding++) {
		/* check encoding is supported on this system/glib */
		gchar *enc_test = g_convert("a", 1, "UTF-8", *encoding, NULL,NULL,NULL);
		if(!enc_test)
			continue;
		g_free(enc_test);

		gtk_combo_box_append_text(GTK_COMBO_BOX(combo_w), *encoding);

		/* check if this is the current encoding */
		if(!g_utf8_collate(*encoding, config_get_s(CFG_NET_ENCODING)))
			have_current_encoding = TRUE;
	}
	if(!have_current_encoding) {
		/* add the default encoding as it was not found in the list */
		gtk_combo_box_prepend_text(GTK_COMBO_BOX(combo_w), config_get_s(CFG_NET_ENCODING));
	}

	return combo_w;
}
#endif

static void
net_type_changed_cb(gint mode, gpointer dlg)
{
	set_config_dlg_modified((struct config_dlg*)dlg, TRUE);
}

static void
net_detect_btn_clicked(GtkButton * button_w, struct config_dlg * dlg)
{
	raise_event(EVENT_IFACE_SHOW_NETDETECT_DLG, NULL, 0);
}

static GtkWidget *
make_net_frame(struct config_dlg * dlg)
{
	GtkWidget * b, * detect_btn;

	b = gtk_vbox_new(FALSE, 2);
	gtk_container_set_border_width(GTK_CONTAINER(b), GEOM_DLG_PAGE_BORDER);

	/* network type option menu */
	dlg->net_type_w = add_widget(
		dlg, b, _("Network connection type:"),
		util_net_type_option(net_type_changed_cb, (gpointer)dlg));

	/* network port numer */
	dlg->net_port_w = add_spin_button(
		dlg, b, CFG_NET_PORT, _("Network port _number:"),
		0, 65535, NULL);

	/* net charset selection widget */
#define ENCODING_FIELD_NAME	_("Network charset:")
#if(WITH_ENCODING_COMBO_WIDGET)
	dlg->net_charset_combo_w = add_widget(
		dlg, b, ENCODING_FIELD_NAME,
		net_frame_encoding_combo(dlg, ENCODING_FIELD_NAME));

	dlg->net_charset_w = GTK_BIN(dlg->net_charset_combo_w)->child;
	g_signal_connect(G_OBJECT(dlg->net_charset_w),
		"changed", G_CALLBACK(entry_changed_cb), (gpointer) dlg);
#else
	dlg->net_charset_w = add_entry(dlg, b, CFG_NET_ENCODING, ENCODING_FIELD_NAME, NULL);
#endif
	/* net masks */
	dlg->net_m1_w = add_entry(
		dlg, b, CFG_NET_MASK1, _("Broadcast mask #1:"),
		&dlg->net_m1_label_w);

	dlg->net_m2_w = add_entry(
		dlg, b, CFG_NET_MASK2, _("Broadcast mask #2:"),
		&dlg->net_m2_label_w);

	/* automatic detection button */
	detect_btn = gtk_button_new_with_mnemonic(_("Detect available networks.."));
	g_signal_connect(G_OBJECT(detect_btn), "clicked",
		G_CALLBACK(net_detect_btn_clicked), (gpointer)dlg);
	add_widget(dlg, b, NULL, detect_btn);

	/* separator for miscelanious settings */
	gtk_box_pack_start(GTK_BOX(b), gtk_hseparator_new(), FALSE, FALSE, 8);

	/* whether to notify buddy/channel of our joins/leaves */
	dlg->net_channel_greet_w = add_check_button(
		b, CFG_CHANNEL_GREET,
		_("Notify channel members when I join or leave"));

	/* whether to reply to info requests */
	dlg->net_reply_w = add_check_button(
		b, CFG_REPLY_INFO, _("_Reply to requests for user information"));

	/* whether to log network events verbosely */
	dlg->net_log_w = add_check_button(
		b, CFG_LOG_NET, _("_Verbosely report network events"));

	gtk_widget_show_all(b);
	return b;
}

static GtkWidget *
make_user_list_frame(struct config_dlg * dlg)
{
	GtkWidget * b;

	b = gtk_vbox_new(FALSE, 2);
	gtk_container_set_border_width(GTK_CONTAINER(b), GEOM_DLG_PAGE_BORDER);

	/* update interval selection */
	dlg->userlist_refresh_w = add_spin_button(
		dlg, b, CFG_USER_LIST_REFRESH, _("Update period in seconds:"),
		1, 1000, NULL);

	/* whether to remove dead users from the list */
	dlg->userlist_keep_unreplied_users_w = add_check_button(
		b, CFG_KEEP_UNREPLIED_USERS,
		_("Keep users that are not replying, on the list"));

	gtk_widget_show_all(b);
	return b;
}

static GtkWidget *
make_auto_away_frame(struct config_dlg * dlg)
{
	GtkWidget * box;

	box = gtk_vbox_new(FALSE, 2);
	gtk_container_set_border_width(GTK_CONTAINER(box), GEOM_DLG_PAGE_BORDER);

	/* insert auto-away & auto-offline */
	dlg->auto_away_w = add_spin_button(
		dlg, box, CFG_AUTO_AWAY, _("Auto-away timeout in minutes:"),
		1, 240, NULL);
	dlg->auto_offline_w = add_spin_button(
		dlg, box, CFG_AUTO_OFFLINE, _("Auto-offline timeout in minutes:"),
		1, 240, NULL);

	gtk_widget_show_all(box);
	return box;
}

static GtkWidget *
make_messages_frame(struct config_dlg * dlg)
{
	GtkWidget * b;

	b = gtk_vbox_new(FALSE, 2);
	gtk_container_set_border_width(GTK_CONTAINER(b), GEOM_DLG_PAGE_BORDER);

	dlg->msg_ignore_mass_w = add_check_button(
		b, PREF_IGNORE_MASS_MSG,
		_("Ignore mass messages"));

	dlg->msg_show_window_w = add_check_button(
		b, CFG_MSG_WINDOW,
		_("Show _messages in popup windows"));

	dlg->msg_max_w = add_spin_button(
		dlg, b, CFG_MAX_MSG,
		_("Max messages on the desktop:"),
		1, 64, &dlg->msg_max_label_w);

	gtk_widget_show_all(b);
	return b;
}

static GtkWidget *
make_pages_frame(struct config_dlg * dlg)
{
	GList * pos_list;
	GtkWidget * b, * w;

	b = gtk_vbox_new(FALSE, 2);
	gtk_container_set_border_width(GTK_CONTAINER(b), GEOM_DLG_PAGE_BORDER);

	/* time-stamp settings */
	dlg->date_channel_w = add_check_button(
		b, CFG_CHANNEL_ST, _("Timestamp c_hannel text"));
	dlg->date_private_w = add_check_button(
		b, CFG_PRIVATE_ST, _("Timestamp _private chat text"));
	dlg->date_status_w = add_check_button(
		b, CFG_STATUS_ST, _("Timestamp s_tatus text"));

	/* add horizontal separator */
	w = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(b), w, FALSE, FALSE, 8);

	/* whether to present (show on top) main window when someone
	 * opens private chat with us */
	dlg->pages_present_on_private = add_check_button(
		b, CFG_PRESENT_ON_PRIVATE,
		_("Present main window when someone opens private chat"));

	/* whether to duplicate new "Status" tab messages
	 * in active sessions */
	dlg->pages_log_current_w = add_check_button(
		b, CFG_LOG_CURRENT,
		_("Duplicate Status messages on the active page"));

	/* add page notebook tab position option
	 * (the order should be the same as in `enum GtkPositionType')
	 */
	pos_list = g_list_append(NULL, _("Left"));
	pos_list = g_list_append(pos_list, _("Right"));
	pos_list = g_list_append(pos_list, _("Top"));
	pos_list = g_list_append(pos_list, _("Bottom"));
	dlg->pages_notebook_pos_w = add_option(
		dlg, b, CFG_NET_TYPE, _("Page tab position:"), pos_list, NULL);
	g_list_free(pos_list);

	gtk_widget_show_all(b);
	return b;
}

static GtkWidget *
make_systray_frame(struct config_dlg * dlg)
{
	GtkWidget * b, * w;
	b = gtk_vbox_new(FALSE, 2);
	gtk_container_set_border_width(GTK_CONTAINER(b), GEOM_DLG_PAGE_BORDER);

	dlg->tray_enable_w = add_check_button(
		b, CFG_TRAY_ICON, _("Enable tray icon"));
#ifdef _WIN32
	gtk_widget_set_sensitive(dlg->tray_enable_w, FALSE);
#endif

	dlg->tray_hide_on_startup_w = add_check_button(
		b, CFG_TRAY_HIDE_ON_STARTUP, _("Hide main window on startup"));

	w = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(b), w, FALSE, FALSE, 8);

	dlg->tray_trigger_joinleave_w = add_check_button(
		b, CFG_TRAY_TRIGGERS, _("Blink icon on Join/Leave events"));
	dlg->tray_trigger_channel_w = add_check_button(
		b, CFG_TRAY_TRIGGERS, _("Blink icon on channel text"));
	dlg->tray_trigger_private_w = add_check_button(
		b, CFG_TRAY_TRIGGERS, _("Blink icon on private chat text"));
	dlg->tray_trigger_status_w = add_check_button(
		b, CFG_TRAY_TRIGGERS, _("Blink icon on Status text"));
	dlg->tray_trigger_topic_w = add_check_button(
		b, CFG_TRAY_TRIGGERS, _("Blink icon on topic change"));

	w = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(b), w, FALSE, FALSE, 8);

	dlg->tray_tooltip_lines_w = add_spin_button(
		dlg, b, CFG_TRAY_TOOLTIP_LINES, _("Max number of lines on the tooltip:"),
		0, 32, & dlg->tray_tooltip_lines_label_w);
		
	gtk_widget_show_all(b);
	return b;
}

static GtkWidget *
make_settings_frame(struct config_dlg * dlg)
{
	GtkWidget * b, * w;

	b = gtk_vbox_new(FALSE, 2);
	gtk_container_set_border_width(GTK_CONTAINER(b), GEOM_DLG_PAGE_BORDER);

	/* add other settings
	 */
	dlg->keep_window_size_w = add_check_button(
		b, CFG_KEEP_WND_SIZE, _("Save & restore window size"));
	dlg->save_on_exit_w = add_check_button(
		b, CFG_AUTO_SAVE, _("_Save settings on exit"));

	w = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(b), w, FALSE, FALSE, 0);

	dlg->motd_w = add_entry(
		dlg, b, CFG_MOTD,
		_("User info text:"), NULL);
	dlg->motd_away_w = add_entry(
		dlg, b, CFG_MOTD_AWAY,
		_("User info text when Away/Offline:"), NULL);

	gtk_widget_show_all(b);
	return b;
}

static gboolean
config_dlg_delete_cb(
	GtkWidget * w,
	GdkEvent * event,
	struct config_dlg * dlg)
{
	gtk_dialog_response(GTK_DIALOG(dlg->dlg_w), GTK_RESPONSE_CANCEL);
	return TRUE;
}

static gboolean
config_dlg_key_cb(
	GtkWidget * dlg_w,
	GdkEventKey * event,
	struct config_dlg * dlg)
{
	/* hide dialog on Esc press */
	if(event->keyval==GDK_Escape) {
		hide_config_dlg(dlg);
		return TRUE;	/* we handle this one, stop propogating */
	}
	return FALSE;	/* propogate event futher */
}

/**
 * config_dlg_response_handler:
 *	handles user's response to the dialog
 */
static void
config_dlg_response_handler(
	GtkDialog * dialog,
	gint response,
	struct config_dlg * dlg)
{
	switch(response) {
	case GTK_RESPONSE_OK:
		if(dlg->is_changed)
			set_config_from_dlg(dlg);

		hide_config_dlg(dlg);
		break;
	case GTK_RESPONSE_CANCEL:
		hide_config_dlg(dlg);
		break;
	case GTK_RESPONSE_APPLY:
		set_config_from_dlg(dlg);
		sync_dlg_from_config(dlg);
		break;
	case GUI_RESPONSE_SAVE:
		set_config_from_dlg(dlg);
		sync_dlg_from_config(dlg);
		raise_event(EVENT_IFACE_STORE_CONFIG, NULL, 0);
		break;
	}	
}

/*
 * stores values from the dialog to the configuration
 */
static void
set_config_from_dlg(struct config_dlg * dlg)
{
	gboolean parsed_ok;
	unsigned long ip;
	unsigned tray_triggers;

	/* signal that we are currently gathering
	 *	configuration and dialog should not be updated
	 *	(scattered) during this time
	 */
	dlg->gathering = TRUE;

	/* Tray icon */
	prefs_set(CFG_TRAY_ICON,
		gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(dlg->tray_enable_w)));
	prefs_set(CFG_TRAY_HIDE_ON_STARTUP,
		gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(dlg->tray_hide_on_startup_w)));

	tray_triggers =
		(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(dlg->tray_trigger_joinleave_w))
				? TRAY_TRIGGER_JOIN_LEAVE: 0)
		| (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(dlg->tray_trigger_channel_w))
				? TRAY_TRIGGER_CHANNEL_TEXT: 0)
		| (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(dlg->tray_trigger_private_w))
				? TRAY_TRIGGER_PRIVATE_TEXT: 0)
		| (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(dlg->tray_trigger_status_w))
				? TRAY_TRIGGER_STATUS_TEXT: 0)
		| (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(dlg->tray_trigger_topic_w))
				? TRAY_TRIGGER_TOPIC_CHANGE: 0);

	prefs_set(CFG_TRAY_TRIGGERS, tray_triggers);
	prefs_set(CFG_TRAY_TOOLTIP_LINES, (unsigned)gtk_spin_button_get_value(
			GTK_SPIN_BUTTON(dlg->tray_tooltip_lines_w)));

	/* Auto away */
	prefs_set(CFG_AUTO_AWAY, (unsigned)gtk_spin_button_get_value(
				GTK_SPIN_BUTTON(dlg->auto_away_w)));
	prefs_set(CFG_AUTO_OFFLINE, (unsigned)gtk_spin_button_get_value(
				GTK_SPIN_BUTTON(dlg->auto_offline_w)));

	/* Pages */
	prefs_set(CFG_PRESENT_ON_PRIVATE, gtk_toggle_button_get_active(
				GTK_TOGGLE_BUTTON(dlg->pages_present_on_private)));
	prefs_set(CFG_LOG_CURRENT, gtk_toggle_button_get_active(
				GTK_TOGGLE_BUTTON(dlg->pages_log_current_w)));
	prefs_set(CFG_PRIVATE_ST, gtk_toggle_button_get_active(
				GTK_TOGGLE_BUTTON(dlg->date_private_w)));
	prefs_set(CFG_CHANNEL_ST, gtk_toggle_button_get_active(
				GTK_TOGGLE_BUTTON(dlg->date_channel_w)));
	prefs_set(CFG_STATUS_ST, gtk_toggle_button_get_active(
				GTK_TOGGLE_BUTTON(dlg->date_status_w)));
	prefs_set(CFG_NOTEBOOK_POS, gtk_option_menu_get_history(
				GTK_OPTION_MENU(dlg->pages_notebook_pos_w)));

	/* Network */
	prefs_set(CFG_NET_TYPE, gtk_option_menu_get_history(GTK_OPTION_MENU(dlg->net_type_w)));
	prefs_set(CFG_NET_PORT, (unsigned)gtk_spin_button_get_value(
				GTK_SPIN_BUTTON(dlg->net_port_w)));
	prefs_set(CFG_NET_ENCODING, gtk_entry_get_text(GTK_ENTRY(dlg->net_charset_w)));
	prefs_set(CFG_USER_LIST_REFRESH, (unsigned)gtk_spin_button_get_value(
				GTK_SPIN_BUTTON(dlg->userlist_refresh_w)));
	prefs_set(CFG_REPLY_INFO, gtk_toggle_button_get_active(
				GTK_TOGGLE_BUTTON(dlg->net_reply_w)));
	prefs_set(CFG_LOG_NET, gtk_toggle_button_get_active(
				GTK_TOGGLE_BUTTON(dlg->net_log_w)));
	prefs_set(CFG_KEEP_UNREPLIED_USERS, gtk_toggle_button_get_active(
				GTK_TOGGLE_BUTTON(dlg->userlist_keep_unreplied_users_w)));
	prefs_set(CFG_CHANNEL_GREET, gtk_toggle_button_get_active(
				GTK_TOGGLE_BUTTON(dlg->net_channel_greet_w)));

	ip = get_ipv4_entry(dlg->net_m1_w, &parsed_ok);
	if(parsed_ok)
		prefs_set(CFG_NET_MASK1, ip);
	set_ipv4_entry(dlg->net_m1_w, config_get_i(CFG_NET_MASK1));
	
	ip = get_ipv4_entry(dlg->net_m2_w, &parsed_ok);
	if(parsed_ok)
		prefs_set(CFG_NET_MASK2, ip);
	set_ipv4_entry(dlg->net_m2_w, config_get_i(CFG_NET_MASK2));

	/* Messages */
	prefs_set(PREF_IGNORE_MASS_MSG, gtk_toggle_button_get_active(
				GTK_TOGGLE_BUTTON(dlg->msg_ignore_mass_w)));
	prefs_set(CFG_MSG_WINDOW, gtk_toggle_button_get_active(
				GTK_TOGGLE_BUTTON(dlg->msg_show_window_w)));
	prefs_set(CFG_MAX_MSG, (unsigned)gtk_spin_button_get_value(
				GTK_SPIN_BUTTON(dlg->msg_max_w)));

	/* Misc */
	prefs_set(CFG_KEEP_WND_SIZE, gtk_toggle_button_get_active(
				GTK_TOGGLE_BUTTON(dlg->keep_window_size_w)));
	prefs_set(CFG_AUTO_SAVE, gtk_toggle_button_get_active(
				GTK_TOGGLE_BUTTON(dlg->save_on_exit_w)));
	prefs_set(CFG_MOTD, gtk_entry_get_text(GTK_ENTRY(dlg->motd_w)));
	prefs_set(CFG_MOTD_AWAY, gtk_entry_get_text(GTK_ENTRY(dlg->motd_away_w)));

	/* change the apply button, to notify that the the dialog is now in sync with config
	 * (however the config was not saved)
	 */
	set_config_dlg_modified(dlg, FALSE);

	/* enable scattering of dialog */
	dlg->gathering = FALSE;
}

/**
 * syncs dialog with the current configuration values
 */
static void
sync_dlg_from_config(struct config_dlg * dlg)
{
	/* Status */
	g_object_set(G_OBJECT(dlg->auto_away_w),
		"value", (gdouble)config_get_i(CFG_AUTO_AWAY), NULL);
	g_object_set(G_OBJECT(dlg->auto_offline_w),
		"value", (gdouble)config_get_i(CFG_AUTO_OFFLINE), NULL);
	
	/* Pages */
	g_object_set(G_OBJECT(dlg->pages_present_on_private),
		"active", (gboolean)config_get_i(CFG_PRESENT_ON_PRIVATE), NULL);
	g_object_set(G_OBJECT(dlg->pages_log_current_w),
		"active", (gboolean)config_get_i(CFG_LOG_CURRENT), NULL);
	g_object_set(G_OBJECT(dlg->date_channel_w),
		"active", (gboolean)config_get_i(CFG_CHANNEL_ST), NULL);
	g_object_set(G_OBJECT(dlg->date_private_w),
		"active", (gboolean)config_get_i(CFG_PRIVATE_ST), NULL);
	g_object_set(G_OBJECT(dlg->date_status_w),
		"active", (gboolean)config_get_i(CFG_STATUS_ST), NULL);
	gtk_option_menu_set_history(
		GTK_OPTION_MENU(dlg->pages_notebook_pos_w),
		config_get_i(CFG_NOTEBOOK_POS));

	/* Misc */
	g_object_set(G_OBJECT(dlg->keep_window_size_w),
		"active", (gboolean)config_get_i(CFG_KEEP_WND_SIZE), NULL);
	g_object_set(G_OBJECT(dlg->save_on_exit_w),
		"active", (gboolean)config_get_i(CFG_AUTO_SAVE), NULL);
	gtk_entry_set_text(GTK_ENTRY(dlg->motd_w), config_get_s(CFG_MOTD));
	gtk_entry_set_text(GTK_ENTRY(dlg->motd_away_w), config_get_s(CFG_MOTD_AWAY));

	/* Tray icon */
	g_object_set(G_OBJECT(dlg->tray_enable_w),
		"active", (gboolean)config_get_i(CFG_TRAY_ICON), NULL);
	g_object_set(G_OBJECT(dlg->tray_hide_on_startup_w),
		"active", (gboolean)config_get_i(CFG_TRAY_HIDE_ON_STARTUP), NULL);
	
	g_object_set(G_OBJECT(dlg->tray_trigger_joinleave_w), "active",
		(gboolean)(config_get_i(CFG_TRAY_TRIGGERS) & TRAY_TRIGGER_JOIN_LEAVE)!=0, NULL);
	g_object_set(G_OBJECT(dlg->tray_trigger_channel_w), "active",
		(gboolean)(config_get_i(CFG_TRAY_TRIGGERS) & TRAY_TRIGGER_CHANNEL_TEXT)!=0, NULL);
	g_object_set(G_OBJECT(dlg->tray_trigger_private_w), "active",
		(gboolean)(config_get_i(CFG_TRAY_TRIGGERS) & TRAY_TRIGGER_PRIVATE_TEXT)!=0, NULL);
	g_object_set(G_OBJECT(dlg->tray_trigger_status_w), "active",
		(gboolean)(config_get_i(CFG_TRAY_TRIGGERS) & TRAY_TRIGGER_STATUS_TEXT)!=0, NULL);
	g_object_set(G_OBJECT(dlg->tray_trigger_topic_w), "active",
		(gboolean)(config_get_i(CFG_TRAY_TRIGGERS) & TRAY_TRIGGER_TOPIC_CHANGE)!=0, NULL);
	cfg_toggle_cb(GTK_TOGGLE_BUTTON(dlg->tray_enable_w), GINT_TO_POINTER(CFG_TRAY_ICON));

	g_object_set(G_OBJECT(dlg->tray_tooltip_lines_w),
			"value", (gdouble)config_get_i(CFG_TRAY_TOOLTIP_LINES), NULL);

	/* Network */
	gtk_option_menu_set_history(
		GTK_OPTION_MENU(dlg->net_type_w),
		config_get_i(CFG_NET_TYPE));
	g_object_set(G_OBJECT(dlg->net_port_w), "value", (gdouble)config_get_i(CFG_NET_PORT), NULL);
	gtk_entry_set_text(GTK_ENTRY(dlg->net_charset_w), config_get_s(CFG_NET_ENCODING));
	g_object_set(G_OBJECT(dlg->net_reply_w),
		"active", (gboolean)config_get_i(CFG_REPLY_INFO), NULL);
	g_object_set(G_OBJECT(dlg->net_log_w),
		"active", (gboolean)config_get_i(CFG_LOG_NET), NULL);
	g_object_set(G_OBJECT(dlg->net_channel_greet_w),
		"active", (gboolean)config_get_i(CFG_CHANNEL_GREET), NULL);
	set_ipv4_entry(dlg->net_m1_w, config_get_i(CFG_NET_MASK1));
	set_ipv4_entry(dlg->net_m2_w, config_get_i(CFG_NET_MASK2));

	/* User list settings */
	g_object_set(G_OBJECT(dlg->userlist_keep_unreplied_users_w),
		"active", (gboolean)config_get_i(CFG_KEEP_UNREPLIED_USERS), NULL);
	g_object_set(G_OBJECT(dlg->userlist_refresh_w),
		"value", (gdouble)config_get_i(CFG_USER_LIST_REFRESH), NULL);

	/* Message settings */
	g_object_set(G_OBJECT(dlg->msg_ignore_mass_w),
		"active", (gboolean)config_get_i(PREF_IGNORE_MASS_MSG), NULL);
	g_object_set(G_OBJECT(dlg->msg_show_window_w),
		"active", (gboolean)config_get_i(CFG_MSG_WINDOW), NULL);
	g_object_set(G_OBJECT(dlg->msg_max_w), "value", (gdouble)config_get_i(CFG_MAX_MSG), NULL);
	cfg_toggle_cb(GTK_TOGGLE_BUTTON(dlg->msg_show_window_w), GINT_TO_POINTER(CFG_MSG_WINDOW));
	
	/* these settings are 'current' - disable the `Save' button */
	set_config_dlg_modified(dlg, FALSE);
}

static struct config_dlg *
make_config_dlg()
{
	GtkWidget * notebook_w;
	struct config_dlg * dlg;

	dlg = g_malloc(sizeof(struct config_dlg));

	dlg->is_changed = FALSE;
	dlg->is_shown = FALSE;
	dlg->gathering = FALSE;

	/* create main window */
	dlg->dlg_w = gtk_dialog_new_with_buttons(
		_("Preferences"), gui_get_main_window(),
		GTK_DIALOG_NO_SEPARATOR, NULL);

	/* add dialog action_area buttons */
	gtk_dialog_add_button(GTK_DIALOG(dlg->dlg_w), GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL);
	dlg->save_w = gtk_dialog_add_button(
			GTK_DIALOG(dlg->dlg_w), GTK_STOCK_SAVE, GUI_RESPONSE_SAVE);
	gtk_button_box_set_child_secondary(
			GTK_BUTTON_BOX(GTK_DIALOG(dlg->dlg_w)->action_area), dlg->save_w, TRUE);
	dlg->apply_w = gtk_dialog_add_button(
			GTK_DIALOG(dlg->dlg_w), GTK_STOCK_APPLY, GTK_RESPONSE_APPLY);
	gtk_dialog_add_button(GTK_DIALOG(dlg->dlg_w), GTK_STOCK_OK, GTK_RESPONSE_OK);

	g_signal_connect(G_OBJECT(dlg->dlg_w), "response",
		G_CALLBACK(config_dlg_response_handler), (gpointer)dlg);
	g_signal_connect(G_OBJECT(dlg->dlg_w), "delete-event",
		G_CALLBACK(config_dlg_delete_cb), (gpointer)dlg);
	g_signal_connect(G_OBJECT(dlg->dlg_w), "key-press-event",
		G_CALLBACK(config_dlg_key_cb), (gpointer)dlg);

	/* setup notebook and insert it into the dialog */
	notebook_w = gtk_notebook_new();
	gtk_widget_show(notebook_w);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg->dlg_w)->vbox), notebook_w, TRUE, TRUE, 0);

	/* add configuration pages */
	gtk_notebook_append_page(
		GTK_NOTEBOOK(notebook_w),
		make_net_frame(dlg), gtk_label_new(_("Network")));
	gtk_notebook_append_page(
		GTK_NOTEBOOK(notebook_w),
		make_user_list_frame(dlg), gtk_label_new(_("User list")));
	gtk_notebook_append_page(
		GTK_NOTEBOOK(notebook_w),
		make_pages_frame(dlg), gtk_label_new(_("Pages")));
	gtk_notebook_append_page(
		GTK_NOTEBOOK(notebook_w),
		make_messages_frame(dlg), gtk_label_new(_("Messages")));
	gtk_notebook_append_page(
		GTK_NOTEBOOK(notebook_w),
		make_systray_frame(dlg), gtk_label_new(_("Tray icon")));
	gtk_notebook_append_page(
		GTK_NOTEBOOK(notebook_w),
		make_auto_away_frame(dlg), gtk_label_new(_("Auto away")));
	gtk_notebook_append_page(
		GTK_NOTEBOOK(notebook_w),
		make_settings_frame(dlg), gtk_label_new(_("Miscellaneous")));

	return dlg;
}

static void
set_config_dlg_modified(struct config_dlg * dlg, gboolean is_changed)
{
	gtk_widget_set_sensitive(dlg->apply_w, is_changed);
	if(is_changed)
		gtk_widget_set_sensitive(dlg->save_w, is_changed);
			
	dlg->is_changed = is_changed;
}

static void
show_config_dlg(struct config_dlg * dlg)
{
	g_assert(dlg && dlg->dlg_w);
	
	gtk_window_present(GTK_WINDOW(dlg->dlg_w));
	
	gtk_widget_set_sensitive(dlg->save_w, !prefs_are_saved());
	dlg->is_shown = TRUE;
}

static void
hide_config_dlg(struct config_dlg * dlg)
{
	g_assert(dlg && dlg->dlg_w);
	gtk_widget_hide(dlg->dlg_w);
	dlg->is_shown = FALSE;
}

static void
destroy_config_dlg(struct config_dlg * dlg)
{
	g_assert(dlg);

	gtk_widget_destroy(dlg->dlg_w);
	g_free(dlg);
}

static void
gui_config_event_cb(
	enum app_event_enum e,
	void * p, int i)
{
	switch(e) {
	case EVENT_MAIN_INIT:
		the_dlg = NULL;
		break;
	case EVENT_MAIN_PRECLOSE:
		if(the_dlg) {
			destroy_config_dlg(the_dlg);
		}
		the_dlg = NULL;
		break;
	case EVENT_CONFIG_CHANGED:
		/* update dialog, if shown and not currently gathering */
		if(the_dlg && the_dlg->is_shown && !the_dlg->gathering) {
			sync_dlg_from_config(the_dlg);
		}
		break;
	case EVENT_CONFIG_SAVED:
		/* disable 'Save' button, as the configuration now is in sync
		 */
		if(the_dlg && the_dlg->is_shown)
			gtk_widget_set_sensitive(the_dlg->save_w, FALSE);
		break;
	case EVENT_IFACE_CONFIGURE_DLG:
		/* make & show dialog */
		if(!the_dlg)
			the_dlg = make_config_dlg();

		sync_dlg_from_config(the_dlg);
		show_config_dlg(the_dlg);
		break;
	default:
		break;	/* nothing */
	}
}

/** exported routines
  *****************************************/
void gui_config_register()
{
	register_event_cb(
		gui_config_event_cb,
		EVENT_MAIN|EVENT_CONFIG|EVENT_IFACE);
}

