/*
 * gui_msg.c: implements message dialog
 * Copyright (C) 2002-2004 Saulius Menkevicius
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>

#include <glib.h>
#include <gtk/gtk.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gdk/gdkkeysyms.h>

#include "main.h"
#include "gui.h"
#include "gui_misc.h"
#include "gui_msg.h"
#include "prefs.h"
#include "user.h"

#define DIALOG_TEXT_WIDTH	300
#define DIALOG_TEXT_HEIGHT	80

#define LOG_MESSAGES	N_("Messages")

struct message_info {
	gpointer remote_user_id;
	gchar * remote_user_name;
	gboolean my;

	GtkWidget	* window_w;
	GtkWidget	* text_w;
	GtkWidget	* send_w, * reply_w, * quote_w;
	GtkTextBuffer	* buffer;
};
#define PMESSAGE_INFO(p) ((struct message_info*)p)

/** static variables
  *****************************************/
static GList * msg_list;
static int msg_count;

/** forward references
 ****************************************/

/** static routines
  ***************************************/

/* message_init:
 *	initializes message list
 */
static void
message_init()
{
	msg_list = NULL;
	msg_count = 0;
}

/* message_delete:
 *	deletes the message and it's window
 *	(if it has any)
 */
static void
message_delete(struct message_info * mi, gboolean destroy_widget)
{
	g_assert(mi && mi->remote_user_name);

	if(destroy_widget) {
		g_assert(mi->window_w);
		gtk_widget_destroy(mi->window_w);
	}

	g_free(mi->remote_user_name);
	g_free(mi);

	/* remove message struct from the list */
	msg_list = g_list_remove(msg_list, mi);
	--msg_count;
}

/**
 * quotes text in received message text (UTF-8)
 * returns:
 *	an allocated GString with quoted message text (UTF-8)
 */
static GString *
message_quote_orig_text(const gchar * orig)
{
	const gchar * prev_c;
	GString * quoted = g_string_new("");

	if(*orig=='\0') {
		return quoted;	/* empty string, -- do nothing */
	}

	g_string_append_c(quoted, '>');

	while(*orig) {
		g_string_append_unichar(quoted, g_utf8_get_char(orig));
		if(*orig=='\n') {
			if(g_utf8_next_char(orig)!='\0') {
				g_string_append_c(quoted, '>');
			}
		}
		prev_c = orig;
		orig = g_utf8_next_char(orig);
	}

	if(*prev_c != '\n') {
		g_string_append_c(quoted, '\n');
	}

	return quoted;
}

/* message_turn_into_my:
 *	changes the message so the user can
 *	edit the text and send it back
 */
static void
message_turn_into_my(struct message_info * mi, gboolean do_quote)
{
	gchar * title;
	gchar * orig;
	GString * quoted = NULL;
	GtkTextIter start, end;

	g_assert(mi->my==FALSE);
	
	mi->my = TRUE;

	if(do_quote) {
		gtk_text_buffer_get_bounds(mi->buffer, &start, &end);
		orig = gtk_text_buffer_get_text(mi->buffer, &start, &end, FALSE);

		quoted = message_quote_orig_text(orig);

		g_free(orig);
	}

	/* preset text of buffer */
	gtk_text_view_set_editable(GTK_TEXT_VIEW(mi->text_w), mi->my);
	gtk_text_buffer_set_text(mi->buffer, quoted ? quoted->str: "", -1);
	if(quoted) {
		g_string_free(quoted, TRUE);
	}

	/* hide reply/quote buttons and emerge send one */
	gtk_widget_hide(mi->reply_w);
	gtk_widget_hide(mi->quote_w);
	gtk_widget_show(mi->send_w);

	/* change title of the message dialog */
	title = g_strdup_printf(_("Message to %s"), mi->remote_user_name);
	gtk_window_set_title(GTK_WINDOW(mi->window_w), title);
	g_free(title);

	/* activate text entry as it might loose cursor when user clicks
	 * on any of dialog buttons */
	gtk_widget_grab_focus(mi->text_w);
}

/* message_cb_key_press_event:
 *	handles dialog key press
 *
 * `Esc' - closes the message
 */
static gboolean
message_cb_key_press_event(
	GtkWidget * window_w,
	GdkEventKey * event,
	struct message_info * mi)
{
	g_assert(mi);

	if(event->keyval==GDK_Escape) {
		/* close this msg */
		message_delete(mi, TRUE);

		/* catch this keypress */
		return TRUE;
	}
	return FALSE;	/* propogate the event futher the chain */
}

/* message_cb_delete_event:
 *	gets notified if the user tries to close the message dialog
 */
static gint
message_cb_delete_event(
	GtkWidget * w, GdkEvent * e,
	struct message_info * mi)
{
	message_delete(mi, FALSE);
	return FALSE;	/* destroy the widget */
}

/* message_cb_response:
 *	handles user response to the dialog:
 *	(cancel/send/reply/close)
 */
static void 
message_cb_response(
	GtkDialog * dialog, gint response,
	struct message_info * mi)
{
	GtkTextIter si, ei;
	gpointer v[2];

	switch(response) {
	case GUI_RESPONSE_QUOTE:
		message_turn_into_my(mi, TRUE);
		break;
	case GUI_RESPONSE_REPLY:
		message_turn_into_my(mi, FALSE);
		break;
	case GUI_RESPONSE_SEND:
		if(mi->remote_user_id) {
			/* "send" */
			gtk_text_buffer_get_start_iter(mi->buffer, &si);
			gtk_text_buffer_get_end_iter(mi->buffer, &ei);

			/* send net message to the user */
			EVENT_V(v,0) = mi->remote_user_id;
			EVENT_V(v,1) = gtk_text_buffer_get_text(
						mi->buffer, &si, &ei, TRUE);
			raise_event(EVENT_MESSAGE_SEND, v, 0);
			g_free(EVENT_V(v,1));

			message_delete(mi, TRUE);
		}
		break;
	case GTK_RESPONSE_CANCEL:
		message_delete(mi, TRUE);
		break;
	}
}


/* message_make_window:
 *	build up and show GTK+ dialog for the message specified (mi)
 */
static void
message_make_window(struct message_info * mi, const gchar * preset_text)
{
	GtkWidget * b, * w, *frame;
	char * title;

	g_assert(mi && mi->remote_user_id && mi->remote_user_name);

	/* make message dialog
	 */
	title = g_strdup_printf(
		mi->my ? _("Message to %s"): _("Message from %s"),
		mi->remote_user_name);

	mi->window_w = gtk_dialog_new_with_buttons(
		title, NULL, GTK_DIALOG_NO_SEPARATOR,
		GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, NULL);
	g_free(title);
	gui_misc_set_icon_for(GTK_WINDOW(mi->window_w));

	/* add send/quote/reply buttons */
	mi->quote_w = gtk_dialog_add_button(
		GTK_DIALOG(mi->window_w),
		GUI_STOCK_QUOTE_MESSAGE, GUI_RESPONSE_QUOTE);
	mi->reply_w = gtk_dialog_add_button(
		GTK_DIALOG(mi->window_w),
		GUI_STOCK_REPLY_MESSAGE, GUI_RESPONSE_REPLY);
	mi->send_w = gtk_dialog_add_button(
		GTK_DIALOG(mi->window_w),
		GUI_STOCK_SEND_MESSAGE, GUI_RESPONSE_SEND);

	if(mi->my) {
		gtk_widget_hide(mi->reply_w);
		gtk_widget_hide(mi->quote_w);
	} else {
		gtk_widget_hide(mi->send_w);
	}
	
	/* register signal source */
	g_signal_connect(
		G_OBJECT(mi->window_w), "delete-event",
		G_CALLBACK(message_cb_delete_event), (gpointer)mi);
	g_signal_connect(
		G_OBJECT(mi->window_w), "key-press-event",
		G_CALLBACK(message_cb_key_press_event), (gpointer)mi);
	g_signal_connect(
		G_OBJECT(mi->window_w), "response",
		G_CALLBACK(message_cb_response), (gpointer)mi);	

	b = gtk_vbox_new(FALSE, 2);
	gtk_widget_show(b);
	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(mi->window_w)->vbox), b);

	frame = gtk_frame_new(_("Message Text"));
	gtk_widget_show(frame);
	gtk_box_pack_start(GTK_BOX(b), frame, TRUE, TRUE, 2);

	w = gtk_scrolled_window_new(NULL, NULL);
	gtk_container_set_border_width(GTK_CONTAINER(w), 4);
	gtk_scrolled_window_set_policy(
		GTK_SCROLLED_WINDOW(w),
		GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_scrolled_window_set_shadow_type(
		GTK_SCROLLED_WINDOW(w),
		GTK_SHADOW_IN);
	gtk_widget_show(w);
	gtk_container_add(GTK_CONTAINER(frame), w);

	mi->text_w = gtk_text_view_new();
	gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(mi->text_w), GTK_WRAP_WORD);
	gtk_widget_set_size_request(mi->text_w, DIALOG_TEXT_WIDTH, DIALOG_TEXT_HEIGHT);
	gtk_widget_show(mi->text_w);
	mi->buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(mi->text_w));
	gtk_text_view_set_border_window_size(
		GTK_TEXT_VIEW(mi->text_w), GTK_TEXT_WINDOW_TOP, 1);
	gtk_text_view_set_border_window_size(
		GTK_TEXT_VIEW(mi->text_w), GTK_TEXT_WINDOW_LEFT, 1);
	gtk_container_add(GTK_CONTAINER(w), mi->text_w);

	/*
	 * Disable `Send' button so the user cant send message if the receiver
	 * cannot or will refuse to receive the message.
	 */
	gtk_widget_set_sensitive(
		mi->send_w,
		IS_MESSAGEABLE_MODE(user_mode_of(mi->remote_user_id)));

	/* preset text of buffer */
	gtk_text_view_set_editable(GTK_TEXT_VIEW(mi->text_w), mi->my);
	if(preset_text!=NULL) {
		gtk_text_buffer_set_text(mi->buffer, preset_text, -1);
	}

	/* show it */
	gtk_window_present(GTK_WINDOW(mi->window_w));
}

/* message_new:
 *	creates a new message and message popup window
 * args:
 *	@user_id	user id of the src/dst of the message
 *	@my		TRUE, if the message is compose one (has 'send' button)
 *	@text		initial text of the message, or NULL
 */
static struct message_info *
message_new(gpointer user_id, gboolean my, const gchar * text)
{
	struct message_info * mi;
	g_assert(user_id);

	/* fillup message struct */
	mi = g_malloc(sizeof(struct message_info));
	mi->remote_user_id = user_id;
	mi->remote_user_name = g_strdup(user_name_of(user_id));
	mi->my = my;

	/* add this message to the list */
	msg_list = g_list_prepend(msg_list, mi);
	++msg_count;

	/* make and show up message window */
	message_make_window(mi, text);

	return mi;
}

/* message_cleanup:
 *	deletes all the messages and all the allocated
 *	structs
 */
static void
message_cleanup()
{
	while(msg_list) {
		message_delete((struct message_info*)msg_list->data, TRUE);
	}
	msg_count = 0;
}

/* message_net_message_ack:
 *	handles remote end confirmation that it has received the message
 */
static void
message_net_message_ack(
	gpointer src_uid,
	const char * text,
	enum user_mode_enum mode)
{
	g_assert(src_uid && text);

	/* just log it */
	if(strlen(text)) {
		log_fevent(
			_(LOG_MESSAGES),
			g_strdup_printf(
				_("%s has received your message (%s)"),
				user_name_of(src_uid), text));
	} else {
		log_fevent(
			_(LOG_MESSAGES),
			g_strdup_printf(
				_("%s has received your message"),
				user_name_of(src_uid)));
	}
}

/* message_net_msg:
 *	shows the message we've received from someone
 */
static void
message_net_msg(gpointer src_user_id, const char * text)
{
	struct message_info * mi;

	if(!config_get_i(CFG_MSG_WINDOW) || msg_count>=config_get_i(CFG_MAX_MSG)) {
		log_fevent(
			_(LOG_MESSAGES),
			g_strdup_printf(
				_("Message from %s: %s"),
				user_name_of(src_user_id), text));
	} else {
		/* create message window */
		mi = message_new(src_user_id, FALSE, text);
	}
}

/* message_user_mode_change:
 *	disables all messages for the specified user,
 *	as he/she gets unreachable
 * args:
 *	@uid		user who changed the mode or became unavailable
 *			(removed from the list)
 *	@new_umode	new user mode (or UMODE_NULL, if the user removed from the list)
 */
static void
message_user_mode_change(
	gpointer uid,
	enum user_mode_enum prev_umode)
{
	GList * l;
	struct message_info * mi;

	g_assert(uid);

	for(l = msg_list; l; l = l->next) {
		mi = PMESSAGE_INFO(l->data);

		if(mi->remote_user_id == uid) {
			if(user_mode_of(uid)==UMODE_NULL) {
				gtk_widget_set_sensitive(mi->send_w, FALSE);
				mi->remote_user_id = NULL;
			} else {
				gtk_widget_set_sensitive(
					mi->send_w, IS_MESSAGEABLE_MODE(user_mode_of(uid)));
			}
		}
	}
}

/* message_user_new:
 *	"revives" some disabled messages if the user became available
 *	after being "dead" for some time
 */
static void
message_user_new(const char * user_name, gpointer new_uid, enum user_mode_enum umode)
{
	GList * l;
	struct message_info * mi;

	g_assert(user_name && new_uid);

	for(l = msg_list; l; l = l->next) {
		mi = PMESSAGE_INFO(l->data);

		if(!mi->remote_user_id && !nickname_cmp(mi->remote_user_name, user_name)) {
			/* ok, msg found - revive it */
			mi->remote_user_id = new_uid;

			/* enable/disable send/reply button */
			if(mi->send_w) {
				gtk_widget_set_sensitive(
					mi->send_w,
					IS_MESSAGEABLE_MODE(umode));
			}
		}
	}
}

/* message_user_rename:
 *	modifies message title, as user changes it's nickname
 */
static void
message_user_rename(gpointer uid, const gchar * new_name)
{
	gchar * title;
	struct message_info * mi;
	GList * l;

	g_assert(uid && new_name);

	for(l = msg_list; l; l = l->next) {
		mi = PMESSAGE_INFO(l->data);
		if(mi->remote_user_id==uid) {
			/* store new nickname */
			g_free(mi->remote_user_name);
			mi->remote_user_name = g_strdup(new_name);

			/* change message dialog title */
			title = g_strdup_printf(
				mi->my ? _("Message to %s"): _("Message from %s"),
				new_name);

			gtk_window_set_title(GTK_WINDOW(mi->window_w), title);
			g_free(title);
		}
	}
}

/* gui_msg_event_cb:
 *	handles app events
 */
static void
gui_msg_event_cb(enum app_event_enum e, void * v, int i)
{
	switch(e) {
	case EVENT_MAIN_INIT:
		message_init();
		break;
	case EVENT_MAIN_PRECLOSE:
		message_cleanup();
		break;
	case EVENT_USER_NEW:
		message_user_new(EVENT_V(v, 1), EVENT_V(v, 0), (enum user_mode_enum)i);
		break;
	case EVENT_USER_REMOVED:
		/* disable messages destined at `v' user */
		message_user_mode_change(v, UMODE_NULL);
		break;
	case EVENT_USER_MODE_CHANGE:
		message_user_mode_change(v, (enum user_mode_enum)i);
		break;
	case EVENT_USER_RENAME:
		message_user_rename(EVENT_V(v, 0), user_name_of(EVENT_V(v, 0)));
		break;
	case EVENT_IFACE_USER_MESSAGE_REQ:
		message_new(v, TRUE, NULL);
		break;
	case EVENT_NET_MSG_MESSAGE:
		message_net_msg(EVENT_V(v, 0), (const char *)EVENT_V(v, 1));
		break;
	case EVENT_NET_MSG_MESSAGE_ACK:
		message_net_message_ack(
			EVENT_V(v, 0),
			(const char*)EVENT_V(v, 1),
			(enum user_mode_enum)i);
		break;
	default:
		break;	/* nothing */
	}
}

/** exported routines
  *****************************************/
void gui_msg_register()
{
	register_event_cb(
		gui_msg_event_cb,
		EVENT_MAIN|EVENT_IFACE|EVENT_USER|EVENT_NET);
}

