/*
 * gui_netdetect_dlg.c: implements network autodetection dialog
 * Copyright (C) 2004 Saulius Menkevicius
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <glib.h>
#include <gtk/gtk.h>
#include <string.h>

#include "main.h"
#include "net.h"
#include "prefs.h"
#include "gui.h"
#include "gui_misc.h"
#include "gui_netselect_dlg.h"

#include "libqcproto/qcproto.h"

#define LIST_MIN_HEIGHT	128

#define DEFAULT_DETECT_RANGE_BEGIN	8167
#define DEFAULT_DETECT_RANGE_END	8169

#define DEFAULT_NET_TYPE		NET_TYPE_VYPRESS

#define DETECT_DURATION	500	/* how much to wait for a response in network */
#define DETECT_STEP	5	/* number of ports to scan in each detection step */
#define DETECT_NICKNAME	"vqcc-net-detect"

struct netselect_dlg {
	GtkWidget * dlg_w,
		* tree_w,
		* detect_begin_spin,
		* detect_end_spin,
		* detect_btn,
		* detect_stop_btn,
		* detect_progress,
		* net_type_option,
		* net_port_spin;
	
	guint detect_timeout_id;
	guint range_begin, range_end, range_current;

	enum net_type_enum detect_type;
	qcs_link detect_links[DETECT_STEP];

	GtkListStore * list;
};

enum network_list_column {
	NETLISTCOL_TYPE_STR,
	NETLISTCOL_PORT_STR,
	NETLISTCOL_TYPE,
	NETLISTCOL_PORT,
	NETLISTCOL_NUM
};

static struct netselect_dlg * main_netselect_dlg;

/* forward ref's */
static void dlg_stop_detection(struct netselect_dlg * dlg, gboolean);
static void netselect_dlg_destroy(struct netselect_dlg * dlg);

static void
detect_range_spin_value_changed(GtkSpinButton * spin_w, struct netselect_dlg * dlg)
{
	gint spin_begin = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(dlg->detect_begin_spin)),
		spin_end = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(dlg->detect_end_spin));

	if((GtkWidget*)spin_w==dlg->detect_begin_spin) {
		/* adjust range end so it doesnt get lower than range begin */
		if(spin_begin > spin_end)
			gtk_spin_button_set_value(
				GTK_SPIN_BUTTON(dlg->detect_end_spin), spin_begin);
	} else {
		/* adjust begining so it doesnt get higher than range end */
		if(spin_end < spin_begin)
			gtk_spin_button_set_value(
				GTK_SPIN_BUTTON(dlg->detect_begin_spin), spin_end);
	}
}

static void
dlg_search_mode(struct netselect_dlg * dlg, gboolean search)
{
	gtk_widget_set_sensitive(dlg->detect_begin_spin, !search);
	gtk_widget_set_sensitive(dlg->detect_end_spin, !search);
	gtk_widget_set_sensitive(dlg->detect_btn, !search);
	gtk_widget_set_sensitive(dlg->detect_stop_btn, search);
}

static void
dlg_add_detected_network(
	struct netselect_dlg * dlg,
	enum net_type_enum type,
	gushort port)
{
	GtkTreeIter iter;
	gchar * port_str;

	gtk_list_store_append(dlg->list, &iter);

	port_str = g_strdup_printf("%u", (guint)port);
	gtk_list_store_set(dlg->list, &iter,
		NETLISTCOL_TYPE_STR, net_name_of_type(type),
		NETLISTCOL_PORT_STR, port_str,
		NETLISTCOL_TYPE, (guint)type,
		NETLISTCOL_PORT, (guint)port,
		-1);
	g_free(port_str);
}

static void
init_detect_links(
	struct netselect_dlg * dlg,
	enum net_type_enum type,
	gushort first_port, gint n_ports)
{
	unsigned long netmasks[2] = { prefs_int(CFG_NET_MASK1), prefs_int(CFG_NET_MASK2) };
	gint i;
	qcs_msg * msg;
       
	/* setup network query msg */
	msg = qcs_newmsg();
	msg->msg = QCS_MSG_REFRESH_REQUEST;
	qcs_msgset(msg, QCS_SRC, DETECT_NICKNAME);
	
	/* open link and send query msg for each link opened */
	for(i = 0; i < n_ports; i++) {
		if(net_connected() && prefs_int(CFG_NET_TYPE)==type
				&& prefs_int(CFG_NET_PORT)==first_port + i) {
			/*
			 * this is the network we're currently connected to;
			 * don't try to connect to it as the listening socket
			 * is already taken by the application
			 */
			dlg->detect_links[i] = NULL;
		} else {
			dlg->detect_links[i] = qcs_open(
				type==NET_TYPE_VYPRESS ? QCS_PROTO_VYPRESS: QCS_PROTO_QCHAT,
				netmasks, first_port + i);

			if(dlg->detect_links[i])
				qcs_send(dlg->detect_links[i], msg);
		}
	}

	qcs_deletemsg(msg);
}

static void
cleanup_detect_links(struct netselect_dlg * dlg)
{
	gint i;
	for(i = 0; i < DETECT_STEP; i++)
		if(dlg->detect_links[i]) {
			qcs_close(dlg->detect_links[i]);
			dlg->detect_links[i] = NULL;
		}
}

static gboolean
link_find_refresh_ack(qcs_link link)
{
	qcs_msg * msg = qcs_newmsg();
	
	while(qcs_waitinput(link, 0) > 0) {
		if(qcs_recv(link, msg)) {
			if(msg->msg==QCS_MSG_REFRESH_ACK
					&& !strcmp(msg->dst, DETECT_NICKNAME)) {
				/* found QCS_MSG_REFRESH_ACK msg for out nickname */
				qcs_deletemsg(msg);
				return TRUE;
			}
		}
	}

	qcs_deletemsg(msg);
	return FALSE;
}

static gboolean
detection_timeout(struct netselect_dlg * dlg)
{
	gfloat progress;
	gint i, range_len;
	
	/* check if any of the links setup in last step had returned with success */
	for(i=0; i < DETECT_STEP; i ++) {
		if(dlg->detect_links[i]) {
		       if(link_find_refresh_ack(dlg->detect_links[i])) {
				/* ok, found at least one refresh ack to our msg */
				dlg_add_detected_network(
					dlg, dlg->detect_type, (gushort)dlg->range_current+i);
		       }
		}
	}
	cleanup_detect_links(dlg);

	/* advance to next network type */
	if(dlg->detect_type < NET_TYPE_LAST) {
		dlg->detect_type ++;
	} else {
		dlg->detect_type = NET_TYPE_FIRST;

		/* advance to next port range */
		dlg->range_current += DETECT_STEP;
		if(dlg->range_current > dlg->range_end) {
			/* ok, reached and of port range and we're finished */
			dlg_stop_detection(dlg, TRUE);
			return FALSE;
		}
	}

	range_len = dlg->range_end - dlg->range_current + 1;
	range_len = range_len < DETECT_STEP ? range_len: DETECT_STEP;

	init_detect_links(dlg, dlg->detect_type, dlg->range_current, range_len);

	/* update progress bar */
	progress = (dlg->range_current - dlg->range_begin) * NET_TYPE_NUM
			+ range_len * dlg->detect_type;
	progress /= (dlg->range_end - dlg->range_begin + 1) * NET_TYPE_NUM;
	gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(dlg->detect_progress), progress);
	
	return TRUE;	/* do not remove me! */
}

static void
dlg_start_detection(struct netselect_dlg * dlg)
{
	gint range_begin, range_end, range_len;

	range_begin = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(dlg->detect_begin_spin));
	range_end = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(dlg->detect_end_spin));

	/* put dialog in search mode (disable some controls) */
	dlg_search_mode(dlg, TRUE);

	/* cleanup current list */
	gtk_list_store_clear(dlg->list);

	if(net_connected()) {
		/* add current network to the list as it won't get detected
		 * due to the socket already held by the application
		 */
		dlg_add_detected_network(dlg, prefs_int(CFG_NET_TYPE), prefs_int(CFG_NET_PORT));
	}

	/* init progress bar */
	gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(dlg->detect_progress), 0.0);

	/* start it */
	dlg->range_begin = range_begin;
	dlg->range_end = range_end;
	dlg->range_current = range_begin;
	dlg->detect_type = NET_TYPE_FIRST;

	memset(&dlg->detect_links, 0, sizeof(qcs_link) * DETECT_STEP);
	range_len = dlg->range_end - dlg->range_current + 1;
	init_detect_links(
		dlg, dlg->detect_type, dlg->range_current,
		range_len > DETECT_STEP ? DETECT_STEP: range_len);

	dlg->detect_timeout_id = g_timeout_add(
		DETECT_DURATION, (GSourceFunc)detection_timeout, (gpointer)dlg);
}

static void
dlg_stop_detection(
	struct netselect_dlg * dlg,
	gboolean select_first)
{
	if(!dlg->detect_timeout_id)
		return;	
	
	/* remove timeout source, if active */
	g_source_remove(dlg->detect_timeout_id);
	dlg->detect_timeout_id = 0;

	/* close any detection qcs_links which left dangling
	 *  if we stoped detection in progress */
	cleanup_detect_links(dlg);
	
	/* put dialog in search mode (disable some controls) */
	dlg_search_mode(dlg, FALSE);

	/* clean progress bar */
	gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(dlg->detect_progress), 0.0);

	if(select_first) {
		GtkTreeIter iter;

		/* select the first entry in the list to be active */
		if(gtk_tree_model_get_iter_first(GTK_TREE_MODEL(dlg->list), &iter)) {
			GtkTreePath * path;

			/* there's at least a single entry in the list, activate the first one */
			path = gtk_tree_model_get_path(GTK_TREE_MODEL(dlg->list), &iter);
			gtk_tree_view_set_cursor(GTK_TREE_VIEW(dlg->tree_w), path, NULL, FALSE);
			gtk_tree_path_free(path);
		}
	}
}

static void
list_row_activated(
	GtkTreeView * treeview_w,
	GtkTreePath * path, GtkTreeViewColumn * column,
	struct netselect_dlg * dlg)
{
	/* send response GTK_RESPONSE_OK,
	 * which will activate the selected network config in dlg_response() */
	gtk_dialog_response(GTK_DIALOG(dlg->dlg_w), GTK_RESPONSE_OK);
}

static void
list_cursor_changed(
	GtkTreeView * treeview_w,
	struct netselect_dlg * dlg)
{
	GtkTreePath * path;

	gtk_tree_view_get_cursor(treeview_w, &path, NULL);
	if(path) {
		GtkTreeIter iter;
		guint net_type, net_port;

		/* get iter to list entry */
		gtk_tree_model_get_iter(GTK_TREE_MODEL(dlg->list), &iter, path);
		gtk_tree_path_free(path);

		/* update manual entry controls */
		gtk_tree_model_get(
			GTK_TREE_MODEL(dlg->list), &iter,
			NETLISTCOL_TYPE, &net_type,
			NETLISTCOL_PORT, &net_port, -1);

		gtk_option_menu_set_history(GTK_OPTION_MENU(dlg->net_type_option), net_type);
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(dlg->net_port_spin), net_port);
	}
}

static void
detect_button_clicked(GtkButton * button_w, struct netselect_dlg * dlg)
{
	if((GtkWidget*)button_w==dlg->detect_btn)
		dlg_start_detection(dlg);
	else
		dlg_stop_detection(dlg, FALSE);
}

static void
dlg_response(GtkDialog * dlg_w, gint response, struct netselect_dlg * dlg)
{
	switch(response) {
	case GTK_RESPONSE_OK:
		/* set network & port */
		prefs_set(CFG_NET_TYPE, gtk_option_menu_get_history(
						GTK_OPTION_MENU(dlg->net_type_option)));
		prefs_set(CFG_NET_PORT, gtk_spin_button_get_value_as_int(
						GTK_SPIN_BUTTON(dlg->net_port_spin)));
		prefs_set(PREF_NET_CONFIGURED, TRUE);

		/* hide the dialog */
		netselect_dlg_destroy(dlg);
		main_netselect_dlg = NULL;
		break;
	case GTK_RESPONSE_CANCEL:
	case GTK_RESPONSE_DELETE_EVENT:
		if(!prefs_int(PREF_NET_CONFIGURED)) {
			/* one cannot get away without network configured :) */
			raise_event(EVENT_IFACE_EXIT, NULL, 0);
		} else {
			/* network is already configured, destroy the dialog and do nothing */
			netselect_dlg_destroy(dlg);
			main_netselect_dlg = NULL;
		}
		break;
	}
}

static void
netselect_dlg_show()
{
	struct netselect_dlg * dlg;
	GtkWidget * main_vbox, * scrolled,
		* button_box, * manual_hbox,
		* hbox, * vbox, * frame;
	GtkTreeViewColumn * column;

	if(main_netselect_dlg) {
		/* dialog is shown already */
		gtk_window_present(GTK_WINDOW(dlg->dlg_w));
		return;
	}
	
	dlg = main_netselect_dlg = g_new(struct netselect_dlg, 1);
	dlg->detect_timeout_id = 0;

	dlg->dlg_w = gtk_dialog_new_with_buttons(
		_("Network detection and configuration"), gui_get_main_window(), 0,
		GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
		GTK_STOCK_OK, GTK_RESPONSE_OK,
		NULL);
	g_signal_connect(G_OBJECT(dlg->dlg_w), "response",
		G_CALLBACK(dlg_response), (gpointer)dlg);
	g_signal_connect(G_OBJECT(dlg->dlg_w), "delete-event", G_CALLBACK(gtk_true), NULL);

	/* upper vertical box for list and port range controls
	 */
	main_vbox = gtk_vbox_new(FALSE, 4);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg->dlg_w)->vbox), main_vbox, TRUE, TRUE, 0);

	frame = gtk_frame_new(_("Detected networks"));
	gtk_container_set_border_width(GTK_CONTAINER(frame), 2);
	gtk_box_pack_start(GTK_BOX(main_vbox), frame, TRUE, TRUE, 0);

	dlg->list = gtk_list_store_new(
		NETLISTCOL_NUM, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_UINT, G_TYPE_UINT);
	dlg->tree_w = gtk_tree_view_new_with_model(GTK_TREE_MODEL(dlg->list));
	g_signal_connect(G_OBJECT(dlg->tree_w), "row-activated",
		G_CALLBACK(list_row_activated), (gpointer)dlg);
	g_signal_connect(G_OBJECT(dlg->tree_w), "cursor-changed",
		G_CALLBACK(list_cursor_changed), (gpointer)dlg);
	
	gtk_tree_view_insert_column_with_attributes(
		GTK_TREE_VIEW(dlg->tree_w), -1, _("Network protocol"),
		gtk_cell_renderer_text_new(), "text", NETLISTCOL_TYPE_STR, NULL);
	column = gtk_tree_view_get_column(GTK_TREE_VIEW(dlg->tree_w), 0);
	gtk_tree_view_column_set_resizable(column, TRUE);
		
	gtk_tree_view_insert_column_with_attributes(
		GTK_TREE_VIEW(dlg->tree_w), -1, _("Port"),
		gtk_cell_renderer_text_new(), "text", NETLISTCOL_PORT_STR, NULL);
	column = gtk_tree_view_get_column(GTK_TREE_VIEW(dlg->tree_w), 1);
	gtk_tree_view_column_set_resizable(column, TRUE);
	
	scrolled = gtk_scrolled_window_new(NULL, NULL);
	gtk_widget_set_size_request(scrolled, -1, LIST_MIN_HEIGHT);
	gtk_scrolled_window_set_policy(
		GTK_SCROLLED_WINDOW(scrolled), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scrolled), dlg->tree_w);
	gtk_container_add(GTK_CONTAINER(frame), scrolled);

	/* Make hbox for lower control frames */
	manual_hbox = gtk_hbox_new(FALSE, 2);
	gtk_box_pack_start(GTK_BOX(main_vbox), manual_hbox, FALSE, FALSE, 0);

	/*
	 * "Automatic scan" frame
	 */
	frame = gtk_frame_new(_("Automatic scan"));
	gtk_container_set_border_width(GTK_CONTAINER(frame), 2);
	gtk_box_pack_start(GTK_BOX(manual_hbox), frame, TRUE, TRUE, 0);

	/* vbox inside the frame */
	vbox = gtk_vbox_new(FALSE, 4);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 4);
	gtk_container_add(GTK_CONTAINER(frame), vbox);

	/* port range start spin
	 */
	hbox = gtk_hbox_new(FALSE, 4);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
	
	dlg->detect_begin_spin = gtk_spin_button_new_with_range(1024, 65535, 1);
	gtk_spin_button_set_value(
		GTK_SPIN_BUTTON(dlg->detect_begin_spin), DEFAULT_DETECT_RANGE_BEGIN);
	gtk_box_pack_end(GTK_BOX(hbox), dlg->detect_begin_spin, FALSE, FALSE, 0);
	gtk_box_pack_end(GTK_BOX(hbox), gtk_label_new(_("Scan UDP ports from")), FALSE, FALSE, 0);

	/* port range end spin
	 */
	hbox = gtk_hbox_new(FALSE, 4);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
	
	dlg->detect_end_spin = gtk_spin_button_new_with_range(1024, 65535, 1);
	gtk_spin_button_set_value(
		GTK_SPIN_BUTTON(dlg->detect_end_spin), DEFAULT_DETECT_RANGE_END);
	gtk_box_pack_end(GTK_BOX(hbox), dlg->detect_end_spin, FALSE, FALSE, 0);
	gtk_box_pack_end(GTK_BOX(hbox), gtk_label_new(_("to (including)")), FALSE, FALSE, 0);

	/* add signal handlers for these spins */
	g_signal_connect(G_OBJECT(dlg->detect_begin_spin), "value-changed",
		G_CALLBACK(detect_range_spin_value_changed), (gpointer)dlg);
	g_signal_connect(G_OBJECT(dlg->detect_end_spin), "value-changed",
		G_CALLBACK(detect_range_spin_value_changed), (gpointer)dlg);

	/* progress bar
	 */
	dlg->detect_progress = gtk_progress_bar_new();
	gtk_box_pack_start(GTK_BOX(vbox), dlg->detect_progress, FALSE, FALSE, 0);

	/* the 'reset' and 'refresh' buttons
	 */
	hbox = gtk_hbox_new(FALSE, 4);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

	button_box = gtk_hbutton_box_new();
	gtk_box_pack_end(GTK_BOX(hbox), button_box, FALSE, FALSE, 0);

	dlg->detect_btn = gtk_button_new_from_stock(GTK_STOCK_REFRESH);
	g_signal_connect(G_OBJECT(dlg->detect_btn), "clicked",
		G_CALLBACK(detect_button_clicked), (gpointer)dlg);
	gtk_box_pack_start(GTK_BOX(button_box), dlg->detect_btn, FALSE, FALSE, 0);

	dlg->detect_stop_btn = gtk_button_new_from_stock(GTK_STOCK_STOP);
	g_signal_connect(G_OBJECT(dlg->detect_stop_btn), "clicked",
		G_CALLBACK(detect_button_clicked), (gpointer)dlg);
	gtk_box_pack_start(GTK_BOX(button_box), dlg->detect_stop_btn, FALSE, FALSE, 0);

	/*
	 * Manual port/network entry frame
	 */
	frame = gtk_frame_new(_("Network settings"));
	gtk_container_set_border_width(GTK_CONTAINER(frame), 2);
	gtk_box_pack_start(GTK_BOX(manual_hbox), frame, TRUE, TRUE, 0);

	/* vbox inside the frame */
	vbox = gtk_vbox_new(FALSE, 4);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 4);
	gtk_container_add(GTK_CONTAINER(frame), vbox);

	/* network type option */
	hbox = gtk_hbox_new(FALSE, 4);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

	dlg->net_type_option = util_net_type_option(NULL, NULL);
	gtk_box_pack_end(GTK_BOX(hbox), dlg->net_type_option, FALSE, FALSE, 0);
	gtk_box_pack_end(GTK_BOX(hbox), gtk_label_new(_("Network type:")), FALSE, FALSE, 0);
	gtk_option_menu_set_history(
		GTK_OPTION_MENU(dlg->net_type_option), prefs_int(CFG_NET_TYPE));

	/* network port number */
	hbox = gtk_hbox_new(FALSE, 4);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

	dlg->net_port_spin = gtk_spin_button_new_with_range(1024, 65535, 1);
	gtk_box_pack_end(GTK_BOX(hbox), dlg->net_port_spin, FALSE, FALSE, 0);
	gtk_box_pack_end(GTK_BOX(hbox), gtk_label_new(_("Port number:")), FALSE, FALSE, 0);
	gtk_spin_button_set_value(
		GTK_SPIN_BUTTON(dlg->net_port_spin),
		prefs_int(CFG_NET_PORT));

	/* show dialog to the user */
	dlg_search_mode(dlg, FALSE);

	if(net_connected()) {
		/* add current network to the list as it won't get detected
		 * due to the socket already held by the application
		 */
		dlg_add_detected_network(dlg, prefs_int(CFG_NET_TYPE), prefs_int(CFG_NET_PORT));
	}

	gtk_widget_show_all(dlg->dlg_w);
}

static void
netselect_dlg_destroy(struct netselect_dlg * dlg)
{
	/* stop any search in progress */
	dlg_stop_detection(dlg, FALSE);

	/* destroy dialog window */
	gtk_widget_destroy(dlg->dlg_w);
	dlg->dlg_w = NULL;

	/* and free any structs ref'd */
	g_object_unref(G_OBJECT(dlg->list));
	dlg->list = NULL;

	g_free(dlg);
}

void netselect_dlg_event(
	enum app_event_enum event,
	gpointer v, gint i)
{
	switch(event) {
	case EVENT_MAIN_INIT:
		main_netselect_dlg = NULL;
		break;
	case EVENT_MAIN_START:
		if(!prefs_int(PREF_NET_CONFIGURED)) {
			g_assert(!main_netselect_dlg);

			netselect_dlg_show();
			dlg_start_detection(main_netselect_dlg);
		}
		break;
	case EVENT_MAIN_PRECLOSE:
		if(main_netselect_dlg) {
			netselect_dlg_destroy(main_netselect_dlg);
			main_netselect_dlg = NULL;
		}
		break;
	case EVENT_IFACE_SHOW_NETDETECT_DLG:
		if(!main_netselect_dlg)
			netselect_dlg_show();
		break;
	case EVENT_NET_ERROR:
		netselect_dlg_show();
		break;
	default: break;
	}
}

void gui_netselect_dlg_register()
{
	register_event_cb(netselect_dlg_event, EVENT_MAIN|EVENT_NET|EVENT_IFACE);
}

