/*
 * gui_tray.c: system tray support for X11/freedesktop.org
 * Copyright (C) 2003-2004 Saulius Menkevicius
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <glib.h>
#include <gtk/gtk.h>
#include <gdk-pixbuf/gdk-pixbuf.h>

#include "misc/eggtrayicon.h"

#include "main.h"
#include "prefs.h"
#include "gui.h"
#include "sess.h"
#include "gui_misc.h"
#include "gui_tray.h"

#include "../pixmaps/traypixbufs.h"

#define ICON_BLINK_TIMEOUT_MS	500
#define TOOLTIP_MAX_LEN		64	/* max chars in one tooltip line */

/** forward references
 */
static void update_tray_icon();
static gboolean create_tray_widget();

/** static variables
 */
static gboolean tray_enabled, tray_embedded;
static GtkWidget * tray_widget;
static GtkWidget * tray_icon_image;
static GdkPixbuf * tray_icon_pixbufs[2];
static guint icon_blink_timeout;
static gboolean tray_blink_state;

static GtkTooltips * tray_tooltips;
static guint tray_tooltip_strings_num;
static GList * tray_tooltip_strings;

static struct misc_menu_info_struct
	tray_menu_descr[] =
{
	{ N_("Normal"),	GUI_STOCK_USER_NORMAL,	TRUE,
			EVENT_IFACE_TRAY_UMODE, (gpointer)UMODE_NORMAL },
	{ N_("Away"),	GUI_STOCK_USER_AWAY,	TRUE,
			EVENT_IFACE_TRAY_UMODE, (gpointer)UMODE_AWAY },
	{ N_("DND"),	GUI_STOCK_USER_DND,	TRUE,
			EVENT_IFACE_TRAY_UMODE, (gpointer)UMODE_DND },
	{ N_("Offline"),GUI_STOCK_USER_OFFLINE,	TRUE,
			EVENT_IFACE_TRAY_UMODE, (gpointer)UMODE_OFFLINE },
	{ N_("Invisible"),GUI_STOCK_USER_INVISIBLE, TRUE,
			EVENT_IFACE_TRAY_UMODE, (gpointer)UMODE_INVISIBLE },
	{ "---", },
	{ N_("Preferences.."), GTK_STOCK_PREFERENCES,TRUE,	EVENT_IFACE_CONFIGURE_DLG, },
	{ "---", },
	{ N_("Quit"),	GTK_STOCK_QUIT,		TRUE,	EVENT_IFACE_EXIT, },
	{ NULL }
};
enum tray_menu_item_enum {
	TRAY_MENU_NORMAL = 0,
	TRAY_MENU_AWAY,
	TRAY_MENU_DND,
	TRAY_MENU_OFFLINE,
	TRAY_MENU_INVISIBLE,
	TRAY_MENU_SEP1,
	TRAY_MENU_PREFERENCES,
	TRAY_MENU_SEP2,
	TRAY_MENU_QUIT
};

/* static routines
 */
static gboolean
tray_blink_cb(gpointer data)
{
	update_tray_icon();
	tray_blink_state = !tray_blink_state;
	
	return TRUE;	/* ie. do not remove this event source */
}

static void
tray_start_blinking()
{
	g_assert(tray_widget);
	if(icon_blink_timeout) return;

	tray_blink_state = TRUE;	/* show the "faded" icon first time it changes */
	icon_blink_timeout = g_timeout_add(ICON_BLINK_TIMEOUT_MS, &tray_blink_cb, NULL);

	/* force icon update */
	tray_blink_cb(NULL);
}

static void
tray_stop_blinking()
{
	g_assert(tray_widget);
	if(icon_blink_timeout==0) return;	/* we're not blinking */
	
	g_source_remove(icon_blink_timeout);
	icon_blink_timeout = 0;

	/* set the icon to "normal" state
	 */
	tray_blink_state = FALSE;
	update_tray_icon();
}

static gboolean
tray_clicked_cb(
	GtkWidget * tray_w,
	GdkEventButton * event,
	void * data)
{
	if(event->type==GDK_BUTTON_PRESS && event->button==1) {
		/* show/hide on left click
		 */
		raise_event(EVENT_IFACE_TRAY_CLICK, 0, 0);
		return TRUE;
	}

	if(event->type==GDK_BUTTON_PRESS && event->button==3) {
		/* menu on right click
		 */
		tray_menu_descr[TRAY_MENU_NORMAL].enabled = my_mode()!=UMODE_NORMAL;
		tray_menu_descr[TRAY_MENU_DND].enabled = my_mode()!=UMODE_DND;
		tray_menu_descr[TRAY_MENU_AWAY].enabled = my_mode()!=UMODE_AWAY;
		tray_menu_descr[TRAY_MENU_OFFLINE].enabled = my_mode()!=UMODE_OFFLINE;
		tray_menu_descr[TRAY_MENU_INVISIBLE].enabled = my_mode()!=UMODE_INVISIBLE;

		misc_popup_menu(tray_menu_descr, event->time, event->button);
		return TRUE;
	}

	return FALSE;
}

static void
tray_embedded_cb(GtkWidget * tray_w, gpointer data)
{
	tray_embedded = TRUE;
	raise_event(EVENT_IFACE_TRAY_EMBEDDED, NULL, 0);
}

static void
tray_destroyed_cb(GtkWidget * tray_w, gpointer data)
{
	tray_widget = NULL;
	tray_embedded = FALSE;
	raise_event(EVENT_IFACE_TRAY_DESTROYED, NULL, 0);

	/* recreate the widget, so it will get shown, once systray area reappears */
	g_idle_add(create_tray_widget, NULL);
}

/**
 * update_tray_icon:
 *	sets tray icon to specified state
 */
static void
update_tray_icon()
{
	if(!tray_widget) return;	/* we have no tray icon here; go away :) */

	gtk_image_set_from_pixbuf(
		GTK_IMAGE(tray_icon_image),
		tray_icon_pixbufs[tray_blink_state ? 1: 0]);
}

static gboolean
create_tray_widget()
{
	GtkWidget * event_box;

	if(tray_widget) {
		gtk_widget_destroy(tray_widget);
		tray_widget = NULL;
	}

	tray_widget = GTK_WIDGET(egg_tray_icon_new("VQCC-gtk"));
	g_object_ref(G_OBJECT(tray_widget));

	event_box = gtk_event_box_new();
	tray_icon_image = gtk_image_new();
	g_signal_connect(G_OBJECT(event_box), "button-press-event", G_CALLBACK(tray_clicked_cb), NULL);
	g_signal_connect(G_OBJECT(tray_widget), "embedded", G_CALLBACK(tray_embedded_cb), NULL);
	g_signal_connect(G_OBJECT(tray_widget), "destroy", G_CALLBACK(tray_destroyed_cb), NULL);
	
	gtk_container_add(GTK_CONTAINER(event_box), tray_icon_image);
	gtk_container_add(GTK_CONTAINER(tray_widget), event_box);
	gtk_widget_show_all(GTK_WIDGET(tray_widget));

	/* sets tray icon image (and size, accordingly) */
	update_tray_icon();

	return FALSE;	/* if we're called as g_idle_add() callback (from tray_destroyed_cb())*/
}

static void
create_tray()
{
	if(tray_enabled) return;

	/* create tray icon pixbufs
	 */
	tray_icon_pixbufs[0] = gdk_pixbuf_new_from_inline(-1, pixbuf_logo_24, FALSE, NULL);
	tray_icon_pixbufs[1] = gdk_pixbuf_new_from_inline(-1, pixbuf_logo_24_act, FALSE, NULL);

	/* tooltip strings */
	tray_tooltip_strings_num = 0;
	tray_tooltip_strings = NULL;
	tray_tooltips = gtk_tooltips_new();
	g_object_ref(G_OBJECT(tray_tooltips));

	icon_blink_timeout = 0;
	tray_embedded = FALSE;

	/* create and embed the widget itself */
	tray_widget = NULL;
	tray_blink_state = FALSE;
	create_tray_widget();	/* this will call update_tray_icon() and set icon image */

	/* set the flag that we have the tray enabled */
	tray_enabled = TRUE;
}

static void
destroy_tray()
{
	if(!tray_enabled) return;

	if(icon_blink_timeout) {
		tray_stop_blinking();
	}

	if(tray_widget) {
		/* remove "destroy" signal handler, so tray_destroyed_cb wouldn't recreate
		 * tray widget again
		 */
		g_signal_handlers_disconnect_by_func(
			G_OBJECT(tray_widget), G_CALLBACK(tray_destroyed_cb), NULL);

		gtk_widget_destroy(tray_widget);
		tray_widget = NULL;	
	}
	
	g_object_unref(G_OBJECT(tray_icon_pixbufs[0]));
	g_object_unref(G_OBJECT(tray_icon_pixbufs[1]));

	/* delete tooltip structs */
	while(tray_tooltip_strings) {
		g_free(tray_tooltip_strings->data);
		tray_tooltip_strings = g_list_delete_link(tray_tooltip_strings, tray_tooltip_strings);
	}
	tray_tooltip_strings_num = 0;
	g_object_unref(tray_tooltips);

	/* misc */
	tray_embedded = FALSE;
	tray_enabled = FALSE;
}

/**
 * tray_blink_trigger:
 *	triggers blinking of tray icon on specific session text events
 */
static void
tray_blink_trigger(
	sess_id session,
	const char * text,
	enum session_text_type text_type)
{
	gboolean blink;

	switch(sess_type(session)) {
	case SESSTYPE_STATUS:
		blink = config_get_i(CFG_TRAY_TRIGGERS) & TRAY_TRIGGER_STATUS_TEXT;
		break;
	case SESSTYPE_PRIVATE:
		blink = (config_get_i(CFG_TRAY_TRIGGERS) & TRAY_TRIGGER_PRIVATE_TEXT
				&& (text_type==SESSTEXT_THEIR_TEXT || text_type==SESSTEXT_THEIR_ME))
			
			|| (config_get_i(CFG_TRAY_TRIGGERS) & TRAY_TRIGGER_JOIN_LEAVE
				&& (text_type==SESSTEXT_JOIN || text_type==SESSTEXT_LEAVE));
		break;
	case SESSTYPE_CHANNEL:
		blink = (config_get_i(CFG_TRAY_TRIGGERS) & TRAY_TRIGGER_TOPIC_CHANGE
				&& text_type==SESSTEXT_TOPIC)
			
			|| (config_get_i(CFG_TRAY_TRIGGERS) & TRAY_TRIGGER_JOIN_LEAVE
				&& (text_type==SESSTEXT_JOIN || text_type==SESSTEXT_LEAVE))

			|| (config_get_i(CFG_TRAY_TRIGGERS) & TRAY_TRIGGER_CHANNEL_TEXT
				&& (text_type==SESSTEXT_THEIR_TEXT || text_type==SESSTEXT_THEIR_ME));
		break;
	}

	if(blink) {
		tray_start_blinking();
	}
}

/**
 * tray_update_tooltip
 *	appends new line to tray icon popup
 */
static void
tray_update_tooltip(gpointer session, const gchar * text)
{
	GList * entry;
	GString * tooltip_text;
	
	g_assert(session && text && tray_tooltips);

	/* don't write the same line twice
	 *  if 'duplicate status messages on the active tab' is enabled
	 */
	if(config_get_i(CFG_LOG_CURRENT)
			&& sess_type(session)==SESSTYPE_STATUS
			&& sess_type(sess_current())!=SESSTYPE_STATUS) {
		return;	
	}

	/* remove old entries so we keep inside specified limit of `CFG_TRAY_TOOLTIP_LINES' */
	while(tray_tooltip_strings
			&& tray_tooltip_strings_num > (config_get_i(CFG_TRAY_TOOLTIP_LINES)-1)) {
		g_free(tray_tooltip_strings->data);
		tray_tooltip_strings = g_list_delete_link(tray_tooltip_strings, tray_tooltip_strings);
		tray_tooltip_strings_num --;
	}

	if(config_get_i(CFG_TRAY_TOOLTIP_LINES)) {
		/* append new line */
		tray_tooltip_strings = g_list_append(tray_tooltip_strings, g_strdup(text));
		tray_tooltip_strings_num ++;

		/* build and set tooltip_strings */
		if(tray_tooltip_strings_num==1) {
			/* it might have been disabled previously (see below, 'else' case) */
			gtk_tooltips_enable(tray_tooltips);
		}

		tooltip_text = g_string_new(NULL);

		for(entry = tray_tooltip_strings; entry; entry = entry->next) {
			/* ensure that tooltip line is not too long */
			gint line_len = g_utf8_strlen((const gchar*)entry->data, -1);
			if(line_len > TOOLTIP_MAX_LEN) {
				/* line is too long, append the beginning of the line only */
				g_string_append_len(
					tooltip_text, (const gchar*)entry->data,
					g_utf8_offset_to_pointer(
							(const gchar*)entry->data, TOOLTIP_MAX_LEN - 2)
						- (const gchar*)entry->data
				);
				g_string_append(tooltip_text, "..");
			} else {			
				g_string_append(tooltip_text, (const gchar*)entry->data);
			}
			if(entry->next)
				g_string_append_c(tooltip_text, '\n');
		}

		/* update tooltip text */
		gtk_tooltips_set_tip(tray_tooltips, tray_widget, tooltip_text->str, NULL);

		g_string_free(tooltip_text, TRUE);
	} else {
		/* disable the tooltip as there are no lines do display
		 */
		gtk_tooltips_disable(tray_tooltips);
	}
}

/**
 * tray_event_cb:
 *	handles application events
 */
static void
tray_event_cb(enum app_event_enum e, gpointer p, int i)
{
	switch(e) {
	case EVENT_MAIN_INIT:
		tray_enabled = FALSE;
		break;
	case EVENT_MAIN_PRECLOSE:
		destroy_tray();
		break;
	case EVENT_CONFIG_CHANGED:
		if((enum config_enum)i == CFG_TRAY_ICON) {
			if(config_get_i(CFG_TRAY_ICON)) {
				create_tray();
			} else {
				destroy_tray();
			}
		}
		break;
	case EVENT_SESSION_TEXT:
		if(tray_widget) {
			tray_update_tooltip(EVENT_V(p, 0), EVENT_V(p, 1));

			if(!gui_is_active())
				tray_blink_trigger(EVENT_V(p, 0), EVENT_V(p, 1), (enum session_text_type)i);
		}
		break;
	case EVENT_IFACE_ACTIVE_CHANGE:
		if(tray_widget && i)
			tray_stop_blinking();
		break;
	case EVENT_IFACE_TRAY_UMODE:
		prefs_set(CFG_MODE, GPOINTER_TO_INT(p));
		break;
	default:
		break;
	}
}

/**
 * gui_tray_embedded:
 *	returns TRUE, if we have icon on system tray visible to the user
 */
gboolean gui_tray_embedded()
{
	return tray_embedded;
}

/**
 * gui_tray_register:
 *	registers gui_tray module for some events
 */
void gui_tray_register()
{
	register_event_cb(
		tray_event_cb,
		EVENT_MAIN|EVENT_CONFIG|EVENT_IFACE|EVENT_SESSION);
}
