/*
 * prefs.c: configuration routines
 * Copyright (C) 2002-2004 Saulius Menkevicius
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdarg.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <glib.h>
#include <gtk/gtk.h>

#include "main.h"
#include "prefs.h"
#include "net.h"
#include "user.h"
#include "gui_tray.h"

#define LOG_PREFS	N_("Preferences")

#define CONFIG_FILE_NAME	".vqcc.conf"
#define CONFIG_IGNORE_FILE_NAME	".vqcc.ignore"

#define DEFAULT_CHARSET		"ISO-8859-1"

enum config_type {
	CFGTYPE_BOOL,
	CFGTYPE_UINT,
	CFGTYPE_STR
};

/** struct defs
  *****************/
struct config_seting_struct {
	const char	* name;
	enum config_type type;
	const char	* default_str;
	unsigned int 	default_i;
	GString	* str;
	unsigned int 	i;
};

/** static vars
  *****************/
static struct config_seting_struct
pref_values[CFG_NUM] = {
	/* !!! these must be in the same order as in `enum config_enum' !!! */
	{ "auto_save",		CFGTYPE_BOOL,	NULL,	TRUE	},
	{ "nickname",		CFGTYPE_STR,	"",	0	},	/* see config_preset() */
	{ "user_list_refresh",	CFGTYPE_UINT,	NULL,	30	},
	{ "mode",		CFGTYPE_UINT,	NULL,	UMODE_NORMAL	},
	{ "net_type",		CFGTYPE_UINT,	NULL,	NET_TYPE_QCHAT	},
	{ "net_port",		CFGTYPE_UINT,	NULL,	8167	},
	{ "net_configured",	CFGTYPE_BOOL,	NULL,	FALSE	},
	{ "log_net",		CFGTYPE_BOOL,	NULL,	TRUE	},
	{ "reply_info",		CFGTYPE_BOOL,	NULL,	TRUE	},
	{ "msg_window",		CFGTYPE_BOOL,	NULL,	TRUE	},
	{ "max_msg",		CFGTYPE_UINT,	NULL,	8	},
	{ "channel_stamp",	CFGTYPE_BOOL,	NULL,	TRUE	},
	{ "private_stamp",	CFGTYPE_BOOL,	NULL,	TRUE	},
	{ "status_stamp",	CFGTYPE_BOOL,	NULL,	FALSE	},
	{ "auto_away",		CFGTYPE_UINT,	NULL,	15	},
	{ "auto_offline",	CFGTYPE_UINT,	NULL,	30	},
	{ "net_mask1",		CFGTYPE_UINT,	NULL,	0	},
	{ "net_mask2",		CFGTYPE_UINT,	NULL,	0	},
	{ "keep_unreplied_users",CFGTYPE_BOOL,	NULL,	FALSE	},
	{ "channel_greet",	CFGTYPE_BOOL,	NULL,	TRUE	},
	{ "log_current",	CFGTYPE_BOOL,	NULL,	TRUE	},
	{ "tray_icon",		CFGTYPE_BOOL,	NULL,	TRUE	},
	{ "notebook_pos",	CFGTYPE_UINT,	NULL,	GTK_POS_TOP },
	{ "popup_timeout",	CFGTYPE_UINT,	NULL,	500	},
	{ "keep_wnd_size",	CFGTYPE_BOOL,	NULL,	FALSE	},
	{ "wnd_size",		CFGTYPE_STR,	"",	0	},
	{ "present_on_private",	CFGTYPE_BOOL,	NULL,	TRUE	},
	{ "tray_triggers",	CFGTYPE_UINT,	NULL,	TRAY_TRIGGER_DEFAULT },
	{ "tray_hide_on_startup",CFGTYPE_BOOL,	NULL,	TRUE	},
	{ "tray_tooltip_lines",	CFGTYPE_UINT,	NULL,	4	},
	{ "wnd_title_max_len",	CFGTYPE_UINT,	NULL,	64	},
	{ "net_encoding",	CFGTYPE_STR,	DEFAULT_CHARSET,0},
	{ "motd",		CFGTYPE_STR,	"",	0	},
	{ "motd_away",		CFGTYPE_STR,	"",	0	},	/* see config_preset() */
	{ "ignore_mass_msg",	CFGTYPE_BOOL,	NULL,	FALSE	},
	{ "persistent_channels",CFGTYPE_STR,	"",	0	}
};
#define cfg pref_values
gboolean pref_values_saved;

static char * config_file_path,
		* ignore_file_path;
static GHashTable * ignore_hash;

/* forward references */
static int config_load();
static int config_store();

/** static routines
  *****************/
static void
emit_all_config()
{
	int i;
	for(i = 0; i < CFG_NUM; i++)
		raise_event(EVENT_CONFIG_CHANGED, NULL, i);
}

static void
config_preset()
{
	int i;
	char * value;
	
	for(i = 0; i < CFG_NUM; i++) {
		switch(cfg[i].type)
		{
		case CFGTYPE_BOOL:
		case CFGTYPE_UINT:
			cfg[i].i = cfg[i].default_i;
			break;
		case CFGTYPE_STR:
			g_assert(cfg[i].default_str);
			cfg[i].str = g_string_new( cfg[i].default_str );
			break;
		}
	}

	/* make random nickname */
	value = g_strdup_printf("guest_%hu", rand()&0xffff);
	g_string_assign(cfg[CFG_NICKNAME].str, value);
	g_free(value);

	/* set default motd */
	g_string_assign(cfg[CFG_MOTD_AWAY].str, _("User is away"));

	/* build ignore list hash */
	ignore_hash = g_hash_table_new_full(
			(GHashFunc)g_str_hash, (GEqualFunc)g_str_equal,
			(GDestroyNotify)g_free, NULL
		);
}

/** config_free:
  *	free any variables we might have allocated in
  *	config module since the start of application
  */
static void
config_free()
{
	int i;
	for(i=0; i < CFG_NUM; i++) {
		if(cfg[i].str) {
			g_string_free(cfg[i].str, TRUE);
			cfg[i].str = NULL;
		}
	}

	if(config_file_path) {
		g_free(config_file_path);
		config_file_path = NULL;
	}
	if(ignore_file_path) {
		g_free(ignore_file_path);
		ignore_file_path = NULL;
	}

	/* destroy ignore list hash table */
	g_hash_table_destroy(ignore_hash);
	ignore_hash = NULL;
}

static void
config_event_cb(
	enum app_event_enum e,
	void * p, int i)
{
	switch(e) {
	case EVENT_MAIN_INIT:
		pref_values_saved = FALSE;
		config_preset();
		break;
	case EVENT_MAIN_START:
		/* reset configuration in all modules */
		emit_all_config();

		/* load "real" configuration */
		config_load();
		break;
	case EVENT_MAIN_CLOSE:
		/* save configuration */
		if(config_get_i(CFG_AUTO_SAVE)) {
			config_store();
		}
		config_free();
		break;

	case EVENT_IFACE_RELOAD_CONFIG:
		/* force reload of config */
		config_load();
		break;

	case EVENT_IFACE_STORE_CONFIG:
		config_store();
		break;
	default:
		break;	/* nothing else */
	}
}

/** extract_equation:
  *	extracts name and value parts of config file equation
  * returns:
  *	1 if successful (both parts extracted)
  *	0 if the line is not equation (empty or comments)
  *	-1 if invalid equation (missing name or value part of)
  */
static int
extract_equation(
	char * string,
	char ** name, char ** value)
{
	char * p;

	g_assert(string && name && value);

	/* check if it's comment */
	p = string;
	while(*p) {
		if(*p=='#') return 0;	/* comments */
		if(*p!=' ' && *p!='\n' && *p!='\t') break; /* non-blank */

		++p;	/* blankspace - skip to next */
	}
	if(!*p) {
		return 0;	/* empty line (with blanks) */
	} else {
		string = p;
	}

	/* ok, found beginning of text, check if it's equation */
	p = strchr(string, '=');
	if(!p || p==string) {
		/* not an equation:
		 *	no '=' or empty `name' */
		return -1;
	}

	/* write '\0' in place of '=',
	 * thus making name a separate asciiz string
	 */
	*p = '\0';

	/* return values */
	*name = string;
	*value = p+1;

	return 1;	/* success */
}

/** get_named_config:
  *	returns configuration enum entry whichs maps
  *	to the config name
  * returns:
  *	CFG_xx if found
  *	CFG_NUM if no such entry found
  */
static enum config_enum
get_named_config(const char * name)
{
	enum config_enum e;

	g_assert(name);

	for(e = 0; e < CFG_NUM; e++) {
		if(!strcmp(cfg[e].name, name)) break;
	}
	return e;
}

/** sets value for specified enum
  *	and emits event for it's change
  * returns:
  *	1 on success
  *	0 on failure (invalid value string)
  */
static int
set_config_value(
	enum config_enum e,
	char * value)
{
	int i;
	g_assert((unsigned int)e<CFG_NUM && value);

	if(cfg[e].type==CFGTYPE_BOOL || cfg[e].type==CFGTYPE_UINT)
	{
		if(!sscanf(value, "%u", &i)) return 0;
		prefs_set(e, i);
	} else {
		/* cfg[e].type==CFGTYPE_STR */
		prefs_set(e, value);
	}
	return 1;	/* success */
}

/** read_config_line:
  *	reads one configuration line (and truncates it to 256 symbols)
  * returns:
  *	non-NULL (char*)
  *		- new line read from file (malloced - caller must free)
  *	NULL	- no more data to be read from file (eof or error reading)
  */
static char *
read_config_line(
	FILE * cf)
{
	char * line, * c;
	int ch;

	g_assert(cf);

	/* allocate line */
	line = g_malloc(256);

	c = line;
	ch = 'd';	/* dummy */
	while(!feof(cf) && !ferror(cf) && (c-line)!=255 && ch!='\n') {
		/* get new char */
		ch = fgetc(cf);
		if(ch==EOF) break;

		/* write char into line */
		*(c++)=(char)ch;
	}
	/* terminate line and remove eoln */
	if(c-line) *(c-(*(c-1)=='\n'))='\0';

	if(c==line) {	/* no symbols read */
		g_free(line);
		return NULL;	/* no more data */
	}

	/** skip any chars that didn't fit into the line of 255 chars */
	if((c-line)==255) {
		while(ch!='\n' && ch!=EOF && !ferror(cf)) {
			fgetc(cf);
		}
	}
	
	return line;
}

/** exported routines
  *******************/

void config_register(
	int argc,
	char ** argv,
	char ** env)
{
	const char * home;
	char ** p;

	register_event_cb(config_event_cb, EVENT_MAIN|EVENT_IFACE);

	/* initialize config_file_path */
	config_file_path = NULL;

	/* check if configuration file was specified */
	for(p=argv+1; *p; p++) {
		if(!g_strcasecmp(*p, "-c") && *(p+1)!=NULL) {
			/* found it, bail out */
			config_file_path = g_strdup(*(p+1));
			break;
		}
	}

	/* get & store UNIX home path */
	home = g_get_home_dir();
	if(!config_file_path) {
		config_file_path = g_strdup_printf(
				"%s/"CONFIG_FILE_NAME, home ? home: "");
	}
	ignore_file_path = g_strdup_printf("%s/"CONFIG_IGNORE_FILE_NAME,
				home ? home: "");
}

/** config_store:
  *	save configuration settings to configuration file
  */
static int config_store()
{
	FILE * cf;
	int cf_num;

	/* open config file */
	cf = fopen(config_file_path, "wb");
	if(!cf) {
		log_ferror(
			_(LOG_PREFS),	
			g_strdup_printf(
				_("Cannot save configuration settings to file \"%s\": %s"),
				config_file_path, strerror(errno))
		);
	} else {
		/* write each setting to file in form
		 *	<config_entry_name> <'='> <config_entry_value> <'\n'>
		 *	if <config_entry_value> is a string, its boundaries
		 *	should be interpreted as start after <'='> and end
		 *	before <'\n'>
		 */
		for(cf_num=0; cf_num<CFG_NUM; cf_num++) {
			switch( cfg[cf_num].type ) {
			case CFGTYPE_BOOL:
				fprintf(cf, "%s=%s\n", cfg[cf_num].name, cfg[cf_num].i ? "1": "0");
				break;
			case CFGTYPE_UINT:
				fprintf(cf, "%s=%u\n", cfg[cf_num].name, cfg[cf_num].i);
				break;
			case CFGTYPE_STR:
				g_assert(cfg[cf_num].str && cfg[cf_num].str->str);
	
				fprintf(cf, "%s=%s\n", cfg[cf_num].name, cfg[cf_num].str->str);
				break;
			}
		}
		fclose(cf);

		log_fevent(
			_(LOG_PREFS),
			g_strdup_printf(
				_("Configuration settings were stored in \"%s\""),
				config_file_path)
		);
	}

	/** store ignore list
	 */
	cf = fopen(ignore_file_path, "w");
	if(cf) {
		GList * ilist = config_ignore_list();

		while(ilist) {
			fprintf(cf, "%s\n", (const char*)ilist->data);
			ilist = g_list_delete_link(ilist, ilist);
		}
		
		fclose(cf);
	} else {
		log_event(
			_(LOG_PREFS),
			g_strdup_printf(
				_("Could not write to ignore list file (\"%s\"): %s"),
				ignore_file_path, strerror(errno))
		);
	}

	/* config values now are in sync with those on the disc */
	pref_values_saved = TRUE;
	raise_event(EVENT_CONFIG_SAVED, NULL, 0);
	
	return 1;	/* success */
}

/** config_load:
  *	load configuration settings from file
  */
static int config_load()
{
	char * ln, * name, * value;
	FILE * cf;
	int e, ln_num;
	enum config_enum config;

	/* parse main file */
	cf = fopen(config_file_path, "r");
	if(!cf) {
		log_ferror(
			_(LOG_PREFS),
			g_strdup_printf(_("Cannot open configuration file \"%s\" for reading: %s"),
				config_file_path, strerror(errno))
		);
		return 0;	/* failure */
	}

	ln_num = 1;
	while((ln=read_config_line(cf))!=NULL) {
		e = extract_equation(ln, &name, &value);
		if(e <= 0) {
			/* error or comment encountered */
			if(e<0) {
				/* invalid equation line */
				log_ferror(
					_(LOG_PREFS),
					g_strdup_printf(
						_("Invalid configuration line \"%s\" (%d)"), ln, ln_num)
				);
			}
		} else {
			/* ok, try to parse name */
			if((config = get_named_config(name))==CFG_NUM) {
				log_ferror(
					_(LOG_PREFS),
					g_strdup_printf(
						_("Invalid configuration entry name \"%s\" on line %d"),
						name, ln_num));
			} else {
				/* set value */
				if(!set_config_value(config, value)) {
					log_ferror(
						_(LOG_PREFS),
						g_strdup_printf(
							_("Invalid configuration value "
								"(\"%s\") for entry \"%s\""),
							value, name)
					);
				}
			}
		}
		g_free(ln);
		++ln_num;
	}
	fclose(cf);

	/* parse ignore list */
	cf = fopen(ignore_file_path, "r");
	if(cf) {
		while((ln=read_config_line(cf))!=NULL) {
			/* add to ignore list if not non-zero file */
			if(strlen(ln)) {
				config_ignore_add(ln);
			}
			g_free(ln);
		}
		fclose(cf);	
	} else {
		log_ferror(
			_(LOG_PREFS),
			g_strdup_printf(
				_("Could not open ignore list file \"%s\": %s"),
				CONFIG_IGNORE_FILE_NAME,
				strerror(errno))
		);
	}

	/* config values are in sync with those on disk */
	pref_values_saved = TRUE;

	return 1;	/* success */
}

gboolean
prefs_are_saved()
{
	return pref_values_saved;
}

void prefs_set(enum config_enum pref, ...)
{
	va_list ap;
	const gchar * str_value;
	gchar * enc_test;
	guint int_value;

	va_start(ap, pref);

	g_return_if_fail(pref < CFG_NUM);

	switch(pref_values[pref].type) {
	case CFGTYPE_BOOL:
	case CFGTYPE_UINT:
		int_value = va_arg(ap, unsigned int);
		break;
	case CFGTYPE_STR:
		str_value = va_arg(ap, const gchar *);
		break;
	}
	
	/* check for cases when some action and/or checking
	 * is required before setting the variable
	 */
	switch(pref) {
	case CFG_NICKNAME:
		if(g_utf8_strlen(str_value, -1)==0) {
			log_error(_(LOG_PREFS), _("Can't set nickname to an empty string"));
			va_end(ap);
			return;
		}
		break;
	case CFG_NET_ENCODING:
		/* check if the encoding specified is usable */
		enc_test = g_convert("a", -1, "UTF-8", str_value, NULL, NULL, NULL);
		if(!enc_test) {
			/* the encoding specified can not be used */
			log_ferror(_(LOG_PREFS), g_strdup_printf(
				_("Charset \"%s\" is invalid or not supported on"
					" this system; faling back to \"%s\""),
				str_value, pref_values[pref].str->str));
			va_end(ap);
			return;
		}
		g_free(enc_test);
		break;
	case CFG_MODE:
		if(int_value >= UMODE_NUM) {
			va_end(ap);
			return;
		}
		break;
	case CFG_NET_TYPE:
		if(int_value >= NET_TYPE_NUM) {
			va_end(ap);
			return;
		}
		break;
	case CFG_NET_PORT:
		if(int_value > 0xffff) {
			va_end(ap);
			return;
		}
		break;
	case CFG_AUTO_AWAY:
		if(int_value==0 || int_value > 240) {	/* invalid time */
			va_end(ap);
			return;
		}

		/* adjust auto offline */
		if(pref_values[CFG_AUTO_OFFLINE].i < int_value)
			prefs_set(CFG_AUTO_OFFLINE, int_value);
		break;
	case CFG_AUTO_OFFLINE:
		if(int_value==0 || int_value > 240) {	/* invalid time */
			va_end(ap);
			return;
		}

		/* adjust auto-away */
		if(pref_values[CFG_AUTO_AWAY].i > int_value)
			prefs_set(CFG_AUTO_AWAY, int_value);
		break;
	default:
		break;
	}

	/* check if current value differs from the previous one
	 * and set it if it differs
	 */
	switch(pref_values[pref].type) {
	case CFGTYPE_UINT:
		if(pref_values[pref].i != int_value) {
			pref_values[pref].i = int_value;
			pref_values_saved = FALSE;
			raise_event(EVENT_CONFIG_CHANGED, NULL, (int)pref);
		}
		break;
	case CFGTYPE_BOOL:
		if((pref_values[pref].i!=0) != (int_value!=0)) {
			pref_values[pref].i = int_value;
			pref_values_saved = FALSE;
			raise_event(EVENT_CONFIG_CHANGED, NULL, (int)pref);
		}
		break;
	case CFGTYPE_STR:
		if(g_utf8_collate(pref_values[pref].str->str, str_value)!=0) {
			g_string_assign(pref_values[pref].str, str_value);
			pref_values_saved = FALSE;
			raise_event(EVENT_CONFIG_CHANGED, NULL, (int)pref);
		}
		break;
	}

	va_end(ap);
}

unsigned int
config_get_i(enum config_enum c)
{
	g_assert(c<CFG_NUM);
	g_assert(cfg[c].type==CFGTYPE_UINT || cfg[c].type==CFGTYPE_BOOL);

#ifdef _WIN32
	/* XXX: tray icon does not function on win32
	 */
	if(c==CFG_TRAY_ICON) {
		return 0;
	}
#endif

	return cfg[c].i;
}

const char *
config_get_s(enum config_enum c)
{
	g_assert(c<CFG_NUM && cfg[c].type==CFGTYPE_STR);
	g_assert(cfg[c].str && cfg[c].str->str);

	return cfg[c].str->str;
}

/** config_is_ignored:
  *	returns non-zero if the user is supposed to be ignored
  */
gboolean config_is_ignored(const gchar * user)
{
	g_assert(user && ignore_hash);

	return g_hash_table_lookup(ignore_hash, user)!=NULL;
}

/** config_ignore_add:
  *	adds specified user to ignore list
  * returns:
  *	zero if already ignored
  */
void config_ignore_add(const gchar * nickname)
{
	g_assert(ignore_hash);

	g_return_if_fail(nickname!=NULL);

	/* check for duplicate */
	if(g_hash_table_lookup(ignore_hash, nickname)!=NULL)
		return;

	/* insert the nickname */
	g_hash_table_insert(
		ignore_hash, g_strdup(nickname),
		GINT_TO_POINTER(1));

	/* inform others about the change */
	raise_event(EVENT_CONFIG_IGNORE_ADDED, (gpointer)nickname, 0);

	/* dump message to status window */
	log_fevent(
		_(LOG_PREFS),
		g_strdup_printf(_("\"%s\" has been inserted into ignore list"), nickname)
	);
}

/** config_ignore_remove:
  *	removes user from ignore list
  * returns:
  *	0 if not found on ignore list
  */
void config_ignore_remove(const gchar * nickname)
{
	gboolean removed;
	g_assert(ignore_hash);

	g_return_if_fail(nickname!=NULL);

	removed = g_hash_table_remove(ignore_hash, nickname);
	if(removed) {
		raise_event(EVENT_CONFIG_IGNORE_REMOVED, (gpointer)nickname, 0);

		/* dump message to status window */
		log_fevent(
			_(LOG_PREFS),
			g_strdup_printf(_("\"%s\" has been removed from ignore list"), nickname)
		);
	}
}

/** config_ignore_list[_cb]:
  *	returns entire ignore list as GList of const char *
  */

static void
config_ignore_list_cb(
	const char * nickname,
	gpointer dummy_value,
	GList ** list)
{
	*list = g_list_prepend(*list, g_strdup(nickname));
}

GList * config_ignore_list()
{
	GList * list = NULL;

	g_assert(ignore_hash);

	/* build GList of (char*)nickname */
	g_hash_table_foreach(
		ignore_hash, (GHFunc)config_ignore_list_cb,
		(gpointer)&list);
	return list;
}


/** config_ignore_remove_all
  *	clears the ignore list
  */
void config_ignore_remove_all()
{
	GList * ignored = config_ignore_list();

	while(ignored) {
		config_ignore_remove((const char*)ignored->data);
		ignored = g_list_delete_link(ignored, ignored);
	}
}
