/*
 * user.c: user list management
 * Copyright (C) 2002-2004 Saulius Menkevicius
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <glib.h>

#include "main.h"
#include "prefs.h"
#include "user.h"

/** struct defs
 **************/
typedef struct user_info_struct {
	GString * name;
	enum user_mode_enum mode;
	gboolean is_active;	/* if user is active (emm.. watching) */
	gboolean is_ignored;	/* if the user is ignored */ 
	gboolean replied;	/* if the user has replied to refresh req */
}
user_info_t;
#define PUSER_INFO_T(p) ((user_info_t*)p)

/** static vars
 **************/
static GHashTable * user_hash;
static gint timeout_left, timeout_value;
static guint timeout_source;

/** static routines
 *********************/
static void set_user_mode(user_info_t *, enum user_mode_enum m);
static void set_user_name(user_info_t *, const gchar * new_name);
static void set_user_active(user_info_t * user, gboolean active);

static void
user_hash_destroy_value(user_info_t * user)
{
	/* sad news.. */
	raise_event(EVENT_USER_REMOVED, user, 0);

	/* release the `user_info_t' struct */
	g_string_free(user->name, TRUE);
	g_free(user);
}

/** remove_dead_cb
 *	removes or marks as dead, users, which did not respond to REFRESH_REQ
 *	(for user with g_hash_table_foreach_remove)
 */
static gboolean
remove_dead_cb(char * key, user_info_t * user, gpointer user_data)
{
	gboolean remove;

	if(!user->replied) {
		/* user didn't reply during specified timeout interval
		 */
		set_user_mode(user, UMODE_DEAD);

		/* remove from the list if configured so */
		remove = !config_get_i(CFG_KEEP_UNREPLIED_USERS);
	} else {
		/* ok, did reply: presume once again as dead
		 */
		user->replied = FALSE;

		/* do not remove */
		remove = FALSE;
	}

	return remove;
}

static void
roll_over_refresh()
{
	/* send user list refresh request */
	raise_event(EVENT_USER_NET_UPDATE_REQ, NULL, 0);

	/* reset timer */
	timeout_left = timeout_value;
}

static gboolean
refresh_timeout_handler(gpointer data)
{
	if(timeout_value) {
		if(--timeout_left==0) {
			/* remove dead users, and reset the rest as `unreplied' */
			g_hash_table_foreach_remove(user_hash, (GHRFunc)remove_dead_cb, NULL);
	
			roll_over_refresh();
		}
	}
	return TRUE;
}

static gpointer
add_user(const gchar * name, enum user_mode_enum mode)
{
	user_info_t * u;
	char * key;
	gpointer event_v[3];

	g_assert(name);

	/* make user_info_t */
	u = g_malloc(sizeof(user_info_t));
	u->name = g_string_new(name);
	u->mode = mode;
	u->is_ignored = FALSE;
	u->is_active = TRUE;
	u->replied = TRUE;	/* assume alive */

	/* make key & check that it's the only one in the list */
	key = g_strdup(name);
	g_assert( g_hash_table_lookup(user_hash, key)==NULL );

	/* insert into the hash table */
	g_hash_table_insert(user_hash, key, u);

	/* send event */
	event_v[0] = (gpointer)u;
	event_v[1] = (gpointer)name;
	event_v[2] = GINT_TO_POINTER(u->is_active);
	raise_event(EVENT_USER_NEW, event_v, (int)mode);

	return u;
}

static gboolean
remove_user_hash_func(gchar * key, user_info_t * user, user_info_t * victim)
{
	if(!victim || user==victim) {
		/* set it to dead mode before removing (gui_page.c uses this to mark user dead) */
		set_user_mode(user, UMODE_DEAD);

		raise_event(EVENT_USER_REMOVED, user, 0);
		return TRUE;
	}
	return FALSE;
}

static void
remove_user(gpointer user)
{
	g_hash_table_foreach_remove(user_hash, (GHRFunc)remove_user_hash_func, user);
}

static void
remove_all_users()
{
	g_hash_table_foreach_remove(user_hash, (GHRFunc)remove_user_hash_func, NULL);
}

static gboolean
mark_all_users_dead_hash_func(gchar * key, user_info_t * user, gpointer dummy)
{
	set_user_mode(user, UMODE_DEAD);
	return FALSE;
}

static void
mark_all_users_dead()
{
	g_hash_table_foreach_remove(
		user_hash,
		config_get_i(CFG_KEEP_UNREPLIED_USERS)
			? (GHRFunc)mark_all_users_dead_hash_func
			: (GHRFunc)remove_user_hash_func,
		NULL
	);
}

static void
set_user_mode(user_info_t * user, enum user_mode_enum new_mode)
{
	if(user->mode != new_mode) {
		enum user_mode_enum prev_mode = user->mode;
		user->mode = new_mode;

		raise_event(EVENT_USER_MODE_CHANGE, user, prev_mode);
	}
}

static void
set_user_active(user_info_t * user, gboolean active)
{
	if(user->is_active != active) {
		user->is_active = active;
		raise_event(EVENT_USER_ACTIVE_CHANGE, (gpointer)user, active);
	}
}

static void
set_user_name(user_info_t * user, const gchar * new_name)
{
	gpointer user_info;
	gpointer key, orig_key, new_key;
	gchar * previous_nick;
	gpointer event[2]; 

	g_assert(user && new_name);

	if(nickname_cmp(user->name->str, new_name)) {
		/* check for nickname collision */
		new_key = g_strdup(new_name);
		if(g_hash_table_lookup(user_hash, new_key)!=NULL) {
			g_free(new_key);

			log_ferror(NULL, g_strdup_printf(
				_("Nickname collision: removing user previously named as %s"),
				user->name->str));
			remove_user(user);
			return;
		}
	
		/* search for original key & struct ptrs */
		key = g_strdup(user->name->str);
		g_hash_table_lookup_extended(user_hash, key, &orig_key, &user_info);

		/* steal key & value from hash table */
		g_hash_table_steal(user_hash, key);

		/* free search & original key */
		g_free(key);
		g_free(orig_key);

		/* insert with the new key */
		g_hash_table_insert(user_hash, new_key, user_info);

		/* set name on struct */
		previous_nick = g_strdup(PUSER_INFO_T(user_info)->name->str);
		g_string_assign(PUSER_INFO_T(user_info)->name, new_name);
	
		/* send update */
		event[0] = user;
		event[1] = previous_nick;
		raise_event(EVENT_USER_RENAME, event, 0);
		g_free(previous_nick);
	}
}

static void
user_event_cb(enum app_event_enum app, void *p, int i)
{
	user_info_t * uinfo;

	switch(app) {
	case EVENT_MAIN_INIT:
		user_hash = g_hash_table_new_full(
			(GHashFunc)g_str_hash,
			(GEqualFunc)g_str_equal,
			(GDestroyNotify)g_free,
			(GDestroyNotify)user_hash_destroy_value);

		/* setup & disable timeout handler */
		timeout_source = g_timeout_add(1000, refresh_timeout_handler, 0);
		timeout_value = 0;
		break;

	case EVENT_MAIN_CLOSE:
		/* destroy hash table (this will also destroy user_info_t's) */
		g_hash_table_destroy(user_hash);

		/* remove timeout event source */
		g_source_remove(timeout_source);
		timeout_source = 0;
		break;

	case EVENT_CONFIG_CHANGED:
		if((enum config_enum)i==CFG_USER_LIST_REFRESH) {
			/* update refresh timeout value
			 */
			timeout_value = config_get_i(CFG_USER_LIST_REFRESH);
			roll_over_refresh();
		}
		break;

	case EVENT_IFACE_USER_REMOVE_REQ:
		remove_user(p);
		break;

	case EVENT_IFACE_USER_LIST_REFRESH_REQ:
		mark_all_users_dead();
		roll_over_refresh();
		break;

	case EVENT_NET_CONNECTED:
		/* enable timeout handler & request user list update */
		timeout_value = config_get_i(CFG_USER_LIST_REFRESH);
		roll_over_refresh();
		break;
	
	case EVENT_NET_DISCONNECTED:
		timeout_value = 0;	/* disable timeout handler */
		remove_all_users();
		break;

	case EVENT_NET_NEW_USER:
		add_user(p, (enum user_mode_enum)i);
		break;

	case EVENT_NET_MSG_CHANNEL_LEAVE:
		if(!channel_cmp("Main", APP_EVENT_V(p, 1))) {
			/* user has left the net: remove from the list */
			if(APP_EVENT_V(p, 0)) {
				remove_user(APP_EVENT_V(p, 0));
			}
		}
		break;
	case EVENT_NET_MSG_REFRESH_ACK:
		/* mark as alive and change/maintain user mode */
		uinfo = PUSER_INFO_T(APP_EVENT_V(p, 0));
		uinfo->replied = TRUE;
		set_user_mode(uinfo, (enum user_mode_enum)i);
		set_user_active(uinfo, GPOINTER_TO_INT(APP_EVENT_V(p, 1)));
		break;

	case EVENT_NET_MSG_RENAME:
		set_user_name(
			PUSER_INFO_T(APP_EVENT_V(p, 0)),
			(const gchar*)APP_EVENT_V(p, 1));
		break;

	case EVENT_NET_MSG_MODE_CHANGE:
		set_user_mode(PUSER_INFO_T(p), (enum user_mode_enum)i);
		break;

	case EVENT_NET_MSG_ACTIVE_CHANGE:
		set_user_active(PUSER_INFO_T(p), i);
		break;

	default:
		break;
	}
}

/** exported routines
 *********************/

void user_register()
{
	register_event_cb(
		user_event_cb,
		EVENT_MAIN|EVENT_CONFIG|EVENT_NET|EVENT_IFACE);
}

gint user_count(void)
{
	g_assert(user_hash);

	return g_hash_table_size(user_hash);
}

void user_list_foreach_func(gpointer key, gpointer user_ptr, GList ** list)
{
	g_assert(list);
	*list = g_list_prepend(*list, user_ptr);
}

GList * user_list()
{
	GList * list = NULL;

	g_hash_table_foreach(user_hash, (GHFunc)user_list_foreach_func, (gpointer)&list);

	return list;
}

gpointer user_by_name(const gchar * name)
{
	g_assert(name);
	return g_hash_table_lookup(user_hash, name);
}

enum user_mode_enum
user_mode_of(gpointer u)
{
	g_assert(u);
	return PUSER_INFO_T(u)->mode;
}

const gchar *
user_name_of(gpointer u)
{
	g_assert(u);
	return PUSER_INFO_T(u)->name->str;
}

void user_set_ignore(gpointer u, int do_ignore)
{
	g_assert(u);
	PUSER_INFO_T(u)->is_ignored = do_ignore!=0;
}

gboolean user_is_ignored(gpointer u)
{
	g_assert(u);
	return PUSER_INFO_T(u)->is_ignored;
}

gboolean user_is_active(gpointer u)
{
	g_assert(u);
	return PUSER_INFO_T(u)->is_active;
}

const gchar *
user_mode_name(enum user_mode_enum local_mode)
{
	const gchar * name;

	switch(local_mode) {
	case UMODE_NORMAL:	name = _("Normal");	break;
	case UMODE_DND:		name = _("DND");	break;
	case UMODE_AWAY:	name = _("Away");	break;
	case UMODE_OFFLINE:	name = _("Offline");	break;
	case UMODE_DEAD:	name = _("*Dead*");	break;
	case UMODE_INVISIBLE:	name = _("Invisible");	break;
	default:
		name = _("(undefined)");
		break;
	}
	return name;
}


